"""
Utilities involved in Packaging.
"""
import contextlib
import functools
import logging
import os
import re
import shutil
import tempfile
import zipfile
from contextlib import contextmanager
from typing import Callable, Dict, List, Optional, cast

import jmespath

from samcli.commands.package.exceptions import ImageNotFoundError, InvalidLocalPathError
from samcli.lib.package.ecr_utils import is_ecr_url
from samcli.lib.package.permissions import (
    AdditiveDirPermissionPermissionMapper,
    AdditiveFilePermissionPermissionMapper,
    PermissionMapper,
    WindowsDirPermissionPermissionMapper,
    WindowsFilePermissionPermissionMapper,
)
from samcli.lib.package.s3_uploader import S3Uploader
from samcli.lib.utils.hash import dir_checksum
from samcli.lib.utils.resources import LAMBDA_LOCAL_RESOURCES

LOG = logging.getLogger(__name__)

# https://docs.aws.amazon.com/AmazonS3/latest/dev-retired/UsingBucket.html
_REGION_PATTERN = r"[a-zA-Z0-9-]+"
_DOT_AMAZONAWS_COM_PATTERN = r"\.amazonaws\.com(\.cn)?"
_S3_URL_REGEXS = [
    # Path-Style (and ipv6 dualstack)
    # - https://s3.Region.amazonaws.com/bucket-name/key name
    # - https://s3.amazonaws.com/bucket-name/key name (old, without region)
    # - https://s3.dualstack.us-west-2.amazonaws.com/...
    re.compile(rf"http(s)?://s3(.dualstack)?(\.{_REGION_PATTERN})?{_DOT_AMAZONAWS_COM_PATTERN}/.+/.+"),
    # Virtual Hosted-Style (including two legacies)
    # https://docs.aws.amazon.com/AmazonS3/latest/userguide/VirtualHosting.html
    # - Virtual Hosted-Style: https://bucket-name.s3.Region.amazonaws.com/key name
    # - Virtual Hosted-Style (Legacy: a dash between S3 and the Region): https://bucket-name.s3-Region.amazonaws.com/...
    # - Virtual Hosted-Style (Legacy Global Endpoint): https://my-bucket.s3.amazonaws.com/...
    re.compile(rf"http(s)?://.+\.s3((.|-){_REGION_PATTERN})?{_DOT_AMAZONAWS_COM_PATTERN}/.+"),
    # S3 access point:
    # - https://AccessPointName-AccountId.s3-accesspoint.region.amazonaws.com
    re.compile(rf"http(s)?://.+-\d+\.s3-accesspoint\.{_REGION_PATTERN}{_DOT_AMAZONAWS_COM_PATTERN}/.+/.+"),
    # S3 protocol URL:
    # - s3://bucket-name/key-name
    re.compile(r"s3://.+/.+"),
]


def is_path_value_valid(path):
    return isinstance(path, str)


def make_abs_path(directory, path):
    if is_path_value_valid(path) and not os.path.isabs(path):
        return os.path.normpath(os.path.join(directory, path))
    return path


def is_s3_protocol_url(url):
    """
    Check whether url is a valid path in the form of "s3://..."
    """
    try:
        S3Uploader.parse_s3_url(url)
        return True
    except ValueError:
        return False


def is_s3_url(url: str) -> bool:
    """
    Check whether a URL is a S3 access URL
    specified at https://docs.aws.amazon.com/AmazonS3/latest/dev-retired/UsingBucket.html
    """
    return any(regex.match(url) for regex in _S3_URL_REGEXS)


def is_local_folder(path):
    return is_path_value_valid(path) and os.path.isdir(path)


def is_local_file(path):
    return is_path_value_valid(path) and os.path.isfile(path)


def is_zip_file(path):
    return is_path_value_valid(path) and zipfile.is_zipfile(path)


def upload_local_image_artifacts(resource_id, resource_dict, property_name, parent_dir, uploader):
    """
    Upload local artifacts referenced by the property at given resource and
    return ECR URL of the uploaded object. It is the responsibility of callers
    to ensure property value is a valid image.

    If path is already a path to S3 object, this method does nothing.

    :param resource_id:     Id of the CloudFormation resource
    :param resource_dict:   Dictionary containing resource definition
    :param property_name:   Property name of CloudFormation resource where this
                            local path is present
    :param parent_dir:      Resolve all relative paths with respect to this
                            directory
    :param uploader:        Method to upload files to ECR

    :return:                ECR URL of the uploaded object
    """

    image_path = jmespath.search(property_name, resource_dict)

    if not image_path:
        message_fmt = "Image not found for {property_name} parameter of {resource_id} resource. \n"
        raise ImageNotFoundError(property_name=property_name, resource_id=resource_id, message_fmt=message_fmt)

    if is_ecr_url(image_path):
        LOG.debug("Property %s of %s is already an ECR URL", property_name, resource_id)
        return image_path

    return uploader.upload(image_path, resource_id)


def upload_local_artifacts(
    resource_type: str,
    resource_id: str,
    resource_dict: Dict,
    property_name: str,
    parent_dir: str,
    uploader: S3Uploader,
    extension: Optional[str] = None,
) -> str:
    """
    Upload local artifacts referenced by the property at given resource and
    return S3 URL of the uploaded object. It is the responsibility of callers
    to ensure property value is a valid string

    If path refers to a file, this method will upload the file. If path refers
    to a folder, this method will zip the folder and upload the zip to S3.
    If path is omitted, this method will zip the current working folder and
    upload.

    If path is already a path to S3 object, this method does nothing.

    :param resource_type:   Type of the CloudFormation resource
    :param resource_id:     Id of the CloudFormation resource
    :param resource_dict:   Dictionary containing resource definition
    :param property_name:   Property name of CloudFormation resource where this
                            local path is present
    :param parent_dir:      Resolve all relative paths with respect to this
                            directory
    :param uploader:        Method to upload files to S3
    :param extension:       Extension of the uploaded artifact
    :return:                S3 URL of the uploaded object
    :raise:                 ValueError if path is not a S3 URL or a local path
    """

    local_path = jmespath.search(property_name, resource_dict)

    if local_path is None:
        # Build the root directory and upload to S3
        local_path = parent_dir

    if is_s3_protocol_url(local_path):
        # A valid CloudFormation template will specify artifacts as S3 URLs.
        # This check is supporting the case where your resource does not
        # refer to local artifacts
        # Nothing to do if property value is an S3 URL
        LOG.debug("Property %s of %s is already a S3 URL", property_name, resource_id)
        return cast(str, local_path)

    local_path = make_abs_path(parent_dir, local_path)

    # Or, pointing to a folder. Zip the folder and upload (zip_method is changed based on resource type)
    if is_local_folder(local_path):
        return zip_and_upload(
            local_path,
            uploader,
            extension,
            zip_method=make_zip_with_lambda_permissions if resource_type in LAMBDA_LOCAL_RESOURCES else make_zip,
        )

    # Path could be pointing to a file. Upload the file
    if is_local_file(local_path):
        return uploader.upload_with_dedup(local_path)

    raise InvalidLocalPathError(resource_id=resource_id, property_name=property_name, local_path=local_path)


def resource_not_packageable(resource_dict):
    inline_code = jmespath.search("InlineCode", resource_dict)
    if inline_code is not None:
        return True
    return False


def zip_and_upload(local_path: str, uploader: S3Uploader, extension: Optional[str], zip_method: Callable) -> str:
    with zip_folder(local_path, zip_method=zip_method) as (zip_file, md5_hash):
        return uploader.upload_with_dedup(zip_file, precomputed_md5=md5_hash, extension=extension)


@contextmanager
def zip_folder(folder_path, zip_method):
    """
    Zip the entire folder and return a file to the zip. Use this inside
    a "with" statement to cleanup the zipfile after it is used.

    Parameters
    ----------
    folder_path : str
        The path of the folder to zip
    zip_method : Callable
        Callable function that takes in a file name and source_path and zips accordingly.

    Yields
    ------
    zipfile_name : str
        Name of the zipfile
    md5hash : str
        The md5 hash of the directory
    """
    md5hash = dir_checksum(folder_path, followlinks=True)
    filename = os.path.join(tempfile.gettempdir(), "data-" + md5hash)

    zipfile_name = zip_method(filename, folder_path)
    try:
        yield zipfile_name, md5hash
    finally:
        if os.path.exists(zipfile_name):
            os.remove(zipfile_name)


def make_zip_with_permissions(file_name, source_root, permission_mappers: List[PermissionMapper]):
    """
    Create a zip file from the source directory

    Parameters
    ----------
    file_name : str
        The basename of the zip file, without .zip
    source_root : str
        The path to the source directory
    permission_mappers : list
        permission objects that need to match an interface such that they have an apply method
        which takes in the external attributes of a zipfile.Zipinfo object
    Returns
    -------
    str
        The name of the zip file, including .zip extension
    """
    permission_mappers = permission_mappers or []
    zipfile_name = "{0}.zip".format(file_name)
    source_root = os.path.abspath(source_root)
    compression_type = zipfile.ZIP_DEFLATED
    with open(zipfile_name, "wb") as f:
        with contextlib.closing(zipfile.ZipFile(f, "w", compression_type)) as zf:
            for root, _, files in os.walk(source_root, followlinks=True):
                for filename in files:
                    full_path = os.path.join(root, filename)
                    relative_path = os.path.relpath(full_path, source_root)
                    with open(full_path, "rb") as data:
                        file_bytes = data.read()
                        info = zipfile.ZipInfo(relative_path)
                        # Context: Nov 2020
                        # Set external attr with Unix 0755 permission
                        # Originally set to 0005 in the discussion below
                        # https://github.com/aws/aws-sam-cli/pull/2193#discussion_r513110608
                        # Changed to 0755 due to a regression in https://github.com/aws/aws-sam-cli/issues/2344
                        # Final PR: https://github.com/aws/aws-sam-cli/pull/2356/files
                        if permission_mappers:
                            # Set host OS to Unix
                            info.create_system = 3
                            # Set current permission of the file/dir to ZipInfo's external_attr
                            info.external_attr = os.stat(full_path).st_mode << 16
                            for permission_mapper in permission_mappers:
                                info = permission_mapper.apply(info)
                            # ZIP date time can be set to the last time the zip content was modified using this logic.
                            # info.date_time = time.localtime()[0:6]

                            # If the date time above is added, the caching logic that compares ZIP files sha will break.
                            # Currently we skip executing sync flows for sam sync command when the logic ZIP hash is
                            # the same as the remote lambda ZIP hash. A timestamp will make the evaluation always false.
                            # However, without this field, contents of the zip file will have a last modified date 1980
                            # because python's zipfile.ZipInfo is set to: https://docs.python.org/3/library/zipfile.html.
                            zf.writestr(info, file_bytes, compress_type=compression_type)
                        else:
                            zf.write(full_path, relative_path)

    return zipfile_name


make_zip = functools.partial(
    make_zip_with_permissions,
    permission_mappers=[
        WindowsFilePermissionPermissionMapper(permissions=0o100755),
        WindowsDirPermissionPermissionMapper(permissions=0o100755),
    ],
)
# Context: Jan 2023
# NOTE(sriram-mv): Modify permissions regardless of the Operating system
# to add 111 for directories and 444 for files in addition to existing permissions.
# No overriding explicit permissions are set.
# Extended Attributes are preserved.
make_zip_with_lambda_permissions = functools.partial(
    make_zip_with_permissions,
    permission_mappers=[
        WindowsFilePermissionPermissionMapper(permissions=0o100755),
        WindowsDirPermissionPermissionMapper(permissions=0o100755),
        AdditiveFilePermissionPermissionMapper(permissions=0o100444),
        AdditiveDirPermissionPermissionMapper(permissions=0o100111),
    ],
)


def copy_to_temp_dir(filepath):
    tmp_dir = tempfile.mkdtemp()
    dst = os.path.join(tmp_dir, os.path.basename(filepath))
    shutil.copyfile(filepath, dst)
    return tmp_dir



"""
Class containing error conditions that are exposed to the user.
"""
import traceback
from typing import IO, Optional
from urllib.parse import quote

import click


class ConfigException(click.ClickException):
    """
    Exception class when configuration file fails checks.
    """


class UserException(click.ClickException):
    """
    Base class for all exceptions that need to be surfaced to the user. Typically, we will display the exception
    message to user and return the error code from CLI process
    """

    exit_code = 1

    def __init__(self, message, wrapped_from=None):
        self.wrapped_from = wrapped_from

        click.ClickException.__init__(self, message)


class UnhandledException(click.ClickException):
    """
    Exception class to re-wrap any exception that is not a UserException.
    Typically this means there is a bug in SAM CLI.
    """

    GH_ISSUE_SEARCH_URL = "https://github.com/aws/aws-sam-cli/issues?q=is%3Aissue+is%3Aopen+{title}"
    GH_BUG_REPORT_URL = "https://github.com/aws/aws-sam-cli/issues/new?template=Bug_report.md&title={title}"
    # NOTE (hawflau): actual exitcode is 1 to not break existing behavior. Only report 255 to telemetry
    exit_code = 1

    def __init__(self, command: str, exception: Exception) -> None:
        self._command = command
        self._exception = exception
        self.__traceback__ = self._exception.__traceback__

        click.ClickException.__init__(self, type(exception).__name__)

    def show(self, file: Optional[IO] = None) -> None:
        """Overriding show to customize printing stack trace and message"""
        if file is None:
            file = click._compat.get_text_stderr()  # pylint: disable=protected-access

        tb = "".join(traceback.format_tb(self.__traceback__))
        click.echo(f"\nError: {self._exception}", file=file, err=True)
        click.echo(f"Traceback:\n{tb}", file=file, err=True)

        encoded_title = quote(f"Bug: {self._command} - {type(self._exception).__name__}")
        lookup_url = self.GH_ISSUE_SEARCH_URL.format(title=encoded_title)
        create_issue_url = self.GH_BUG_REPORT_URL.format(title=encoded_title)
        msg = (
            f'An unexpected error was encountered while executing "{self._command}".\n'
            "Search for an existing issue:\n"
            f"{lookup_url}\n"
            "Or create a bug report:\n"
            f"{create_issue_url}"
        )
        click.secho(msg, file=file, err=True, fg="yellow")


class CredentialsError(UserException):
    """
    Exception class when credentials that have been passed are invalid.
    """


class SDKError(UserException):
    """
    Exception class when there are generic Boto Errors.
    """


class SchemasApiException(UserException):
    """
    Exception class to wrap all Schemas APIs exceptions.
    """


class RegionError(UserException):
    """
    Exception class when no valid region is passed to a client.
    """


class AppTemplateUpdateException(UserException):
    """
    Exception class when updates to app templates for init enters an unstable state.
    """


class LambdaImagesTemplateException(UserException):
    """
    Exception class when multiple Lambda Image app templates are found for any runtime
    """


class ContainersInitializationException(UserException):
    """
    Exception class when SAM is not able to initialize any of the lambda functions containers
    """


class PipelineTemplateCloneException(UserException):
    """
    Exception class when unable to download pipeline templates from a Git repository during `sam pipeline init`
    """


class AppPipelineTemplateManifestException(UserException):
    """
    Exception class when SAM is not able to parse the "manifest.yaml" file located in the SAM pipeline templates
    Git repo: "github.com/aws/aws-sam-cli-pipeline-init-templates.git
    """


class AppPipelineTemplateMetadataException(UserException):
    """
    Exception class when SAM is not able to parse the "metadata.json" file located in the SAM pipeline templates
    """


class InvalidInitOptionException(UserException):
    """
    Exception class when user provides wrong options
    """


class InvalidImageException(UserException):
    """
    Value provided to --build-image or --invoke-image is invalid URI
    """


class InvalidStackNameException(UserException):
    """
    Value provided to --stack-name is invalid
    """


class LinterRuleMatchedException(UserException):
    """
    The linter matched a rule meaning that the template linting failed
    """



"""
Logic for uploading to s3 based on supplied template file and s3 bucket
"""

# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import logging
import os
from typing import Optional

import boto3
import click
import docker

from samcli.commands.package.exceptions import PackageFailedError
from samcli.lib.intrinsic_resolver.intrinsics_symbol_table import IntrinsicsSymbolTable
from samcli.lib.package.artifact_exporter import Template
from samcli.lib.package.code_signer import CodeSigner
from samcli.lib.package.ecr_uploader import ECRUploader
from samcli.lib.package.s3_uploader import S3Uploader
from samcli.lib.package.uploaders import Uploaders
from samcli.lib.providers.provider import ResourceIdentifier, get_resource_full_path_by_id
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider
from samcli.lib.utils.boto_utils import get_boto_config_with_user_agent
from samcli.yamlhelper import yaml_dump

LOG = logging.getLogger(__name__)


class PackageContext:

    MSG_PACKAGED_TEMPLATE_WRITTEN = (
        "\nSuccessfully packaged artifacts and wrote output template "
        "to file {output_file_name}."
        "\n"
        "Execute the following command to deploy the packaged template"
        "\n"
        "sam deploy --template-file {output_file_path} "
        "--stack-name <YOUR STACK NAME>"
        "\n"
    )

    uploaders: Uploaders

    def __init__(
        self,
        template_file,
        s3_bucket,
        image_repository,
        image_repositories,
        s3_prefix,
        kms_key_id,
        output_template_file,
        use_json,
        force_upload,
        no_progressbar,
        metadata,
        region,
        profile,
        on_deploy=False,
        signing_profiles=None,
    ):
        self.template_file = template_file
        self.s3_bucket = s3_bucket
        self.image_repository = image_repository
        self.image_repositories = image_repositories
        self.s3_prefix = s3_prefix
        self.kms_key_id = kms_key_id
        self.output_template_file = output_template_file
        self.use_json = use_json
        self.force_upload = force_upload
        self.no_progressbar = no_progressbar
        self.metadata = metadata
        self.region = region
        self.profile = profile
        self.on_deploy = on_deploy
        self.code_signer = None
        self.signing_profiles = signing_profiles
        self._global_parameter_overrides = {IntrinsicsSymbolTable.AWS_REGION: region} if region else {}

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def run(self):
        """
        Execute packaging based on the argument provided by customers and samconfig.toml.
        """
        stacks, _ = SamLocalStackProvider.get_stacks(
            self.template_file,
            global_parameter_overrides=self._global_parameter_overrides,
        )
        self.image_repositories = self.image_repositories if self.image_repositories is not None else {}
        updated_repo = {}
        for image_repo_func_id, image_repo_uri in self.image_repositories.items():
            repo_full_path = get_resource_full_path_by_id(stacks, ResourceIdentifier(image_repo_func_id))
            if repo_full_path:
                updated_repo[repo_full_path] = image_repo_uri
        self.image_repositories = updated_repo
        region_name = self.region if self.region else None

        s3_client = boto3.client(
            "s3",
            config=get_boto_config_with_user_agent(signature_version="s3v4", region_name=region_name),
        )
        ecr_client = boto3.client("ecr", config=get_boto_config_with_user_agent(region_name=region_name))

        docker_client = docker.from_env()

        s3_uploader = S3Uploader(
            s3_client, self.s3_bucket, self.s3_prefix, self.kms_key_id, self.force_upload, self.no_progressbar
        )
        # attach the given metadata to the artifacts to be uploaded
        s3_uploader.artifact_metadata = self.metadata
        ecr_uploader = ECRUploader(
            docker_client, ecr_client, self.image_repository, self.image_repositories, self.no_progressbar
        )

        self.uploaders = Uploaders(s3_uploader, ecr_uploader)

        code_signer_client = boto3.client("signer", config=get_boto_config_with_user_agent(region_name=region_name))
        self.code_signer = CodeSigner(code_signer_client, self.signing_profiles)

        try:
            exported_str = self._export(self.template_file, self.use_json)

            self.write_output(self.output_template_file, exported_str)

            if self.output_template_file and not self.on_deploy:
                msg = self.MSG_PACKAGED_TEMPLATE_WRITTEN.format(
                    output_file_name=self.output_template_file,
                    output_file_path=os.path.abspath(self.output_template_file),
                )
                click.echo(msg)
        except OSError as ex:
            raise PackageFailedError(template_file=self.template_file, ex=str(ex)) from ex

    def _export(self, template_path, use_json):
        template = Template(
            template_path,
            os.getcwd(),
            self.uploaders,
            self.code_signer,
            normalize_template=True,
            normalize_parameters=True,
        )
        exported_template = template.export()

        if use_json:
            exported_str = json.dumps(exported_template, indent=4, ensure_ascii=False)
        else:
            exported_str = yaml_dump(exported_template)

        return exported_str

    @staticmethod
    def write_output(output_file_name: Optional[str], data: str) -> None:
        if output_file_name is None:
            click.echo(data)
            return

        with open(output_file_name, "w") as fp:
            fp.write(data)



"""
HandlerObserver and its helper classes.
"""
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, List, Optional

from watchdog.events import (
    FileSystemEvent,
    FileSystemEventHandler,
    RegexMatchingEventHandler,
)
from watchdog.observers import Observer
from watchdog.observers.api import DEFAULT_OBSERVER_TIMEOUT, ObservedWatch


@dataclass
class PathHandler:
    """PathHandler is an object that can be passed into
    Bundle Observer directly for watching a specific path with
    corresponding EventHandler

    Fields:
        event_handler : FileSystemEventHandler
            Handler for the event
        path : Path
            Path to the folder to be watched
        recursive : bool, optional
            True to watch child folders, by default False
        static_folder : bool, optional
            Should the observed folder name be static, by default False
            See StaticFolderWrapper on the use case.
        self_create : Optional[Callable[[], None]], optional
            Callback when the folder to be observed itself is created, by default None
            This will not be called if static_folder is False
        self_delete : Optional[Callable[[], None]], optional
            Callback when the folder to be observed itself is deleted, by default None
            This will not be called if static_folder is False
    """

    event_handler: FileSystemEventHandler
    path: Path
    recursive: bool = False
    static_folder: bool = False
    self_create: Optional[Callable[[], None]] = None
    self_delete: Optional[Callable[[], None]] = None


class StaticFolderWrapper:
    """This class is used to alter the behavior of watchdog folder watches.
    https://github.com/gorakhargosh/watchdog/issues/415
    By default, if a folder is renamed, the handler will still get triggered for the new folder
    Ex:
        1. Create FolderA
        2. Watch FolderA
        3. Rename FolderA to FolderB
        4. Add file to FolderB
        5. Handler will get event for adding the file to FolderB but with event path still as FolderA
    This class watches the parent folder and if the folder to be watched gets renamed or deleted,
    the watch will be stopped and changes in the renamed folder will not be triggered.
    """

    def __init__(self, observer: "HandlerObserver", initial_watch: ObservedWatch, path_handler: PathHandler):
        """[summary]

        Parameters
        ----------
        observer : HandlerObserver
            HandlerObserver
        initial_watch : ObservedWatch
            Initial watch for the folder to be watched that gets returned by HandlerObserver
        path_handler : PathHandler
            PathHandler of the folder to be watched.
        """
        self._observer = observer
        self._path_handler = path_handler
        self._watch = initial_watch

    def _on_parent_change(self, _: FileSystemEvent) -> None:
        """Callback for changes detected in the parent folder"""

        # When folder is being watched but the folder does not exist
        if self._watch and not self._path_handler.path.exists():
            if self._path_handler.self_delete:
                self._path_handler.self_delete()
            self._observer.unschedule(self._watch)
            self._watch = None
        # When folder is not being watched but the folder does exist
        elif not self._watch and self._path_handler.path.exists():
            if self._path_handler.self_create:
                self._path_handler.self_create()
            self._watch = self._observer.schedule_handler(self._path_handler)

    def get_dir_parent_path_handler(self) -> PathHandler:
        """Get PathHandler that watches the folder changes from the parent folder.

        Returns
        -------
        PathHandler
            PathHandler for the parent folder. This should be added back into the HandlerObserver.
        """
        dir_path = self._path_handler.path.resolve()
        parent_dir_path = dir_path.parent
        parent_folder_handler = RegexMatchingEventHandler(
            regexes=[f"^{re.escape(str(dir_path))}$"],
            ignore_regexes=[],
            ignore_directories=False,
            case_sensitive=True,
        )
        parent_folder_handler.on_any_event = self._on_parent_change
        return PathHandler(path=parent_dir_path, event_handler=parent_folder_handler)


class HandlerObserver(Observer):  # pylint: disable=too-many-ancestors
    """
    Extended WatchDog Observer that takes in a single PathHandler object.
    """

    def __init__(self, timeout=DEFAULT_OBSERVER_TIMEOUT):
        super().__init__(timeout=timeout)

    def schedule_handlers(self, path_handlers: List[PathHandler]) -> List[ObservedWatch]:
        """Schedule a list of PathHandlers

        Parameters
        ----------
        path_handlers : List[PathHandler]
            List of PathHandlers to be scheduled

        Returns
        -------
        List[ObservedWatch]
            List of ObservedWatch corresponding to path_handlers in the same order.
        """
        watches = list()
        for path_handler in path_handlers:
            watches.append(self.schedule_handler(path_handler))
        return watches

    def schedule_handler(self, path_handler: PathHandler) -> ObservedWatch:
        """Schedule a PathHandler

        Parameters
        ----------
        path_handler : PathHandler
            PathHandler to be scheduled

        Returns
        -------
        ObservedWatch
            ObservedWatch corresponding to the PathHandler.
            If static_folder is True, the parent folder watch will be returned instead.
        """
        watch = self.schedule(path_handler.event_handler, str(path_handler.path), path_handler.recursive)
        if path_handler.static_folder:
            static_wrapper = StaticFolderWrapper(self, watch, path_handler)
            parent_path_handler = static_wrapper.get_dir_parent_path_handler()
            watch = self.schedule_handler(parent_path_handler)
        return watch



"""
The container for Endpoints
"""
from dataclasses import dataclass
from typing import Any


@dataclass
class EndpointsDef:
    """
    Dataclass for containing entries of endpoints data
    """

    LogicalResourceId: str
    PhysicalResourceId: str
    CloudEndpoint: Any
    Methods: Any



"""
build utilities
"""
import logging
from typing import Union, Dict, Optional

from samcli.commands.local.lib.exceptions import OverridesNotWellDefinedError
from samcli.lib.providers.provider import Function, LayerVersion

LOG = logging.getLogger(__name__)


def _make_env_vars(
    resource: Union[Function, LayerVersion], file_env_vars: Dict, inline_env_vars: Optional[Dict]
) -> Dict:
    """Returns the environment variables configuration for this function

    Priority order (high to low):
    1. Function specific env vars from command line
    2. Function specific env vars from json file
    3. Global env vars from command line
    4. Global env vars from json file

    Parameters
    ----------
    resource : Union[Function, LayerVersion]
        Lambda function or layer to generate the configuration for
    file_env_vars : Dict
        The dictionary of environment variables loaded from the file
    inline_env_vars : Optional[Dict]
        The optional dictionary of environment variables defined inline

    Returns
    -------
    dictionary
        Environment variable configuration for this function

    Raises
    ------
    samcli.commands.local.lib.exceptions.OverridesNotWellDefinedError
        If the environment dict is in the wrong format to process environment vars

    """

    name = resource.name
    result = {}

    # validate and raise OverridesNotWellDefinedError
    for env_var in list((file_env_vars or {}).values()) + list((inline_env_vars or {}).values()):
        if not isinstance(env_var, dict):
            reason = "Environment variables {} in incorrect format".format(env_var)
            LOG.debug(reason)
            raise OverridesNotWellDefinedError(reason)

    if file_env_vars:
        parameter_result = file_env_vars.get("Parameters", {})
        result.update(parameter_result)

    if inline_env_vars:
        inline_parameter_result = inline_env_vars.get("Parameters", {})
        result.update(inline_parameter_result)

    if file_env_vars:
        specific_result = file_env_vars.get(name, {})
        result.update(specific_result)

    if inline_env_vars:
        inline_specific_result = inline_env_vars.get(name, {})
        result.update(inline_specific_result)

    return result



"""
Cookiecutter-based generation logic for project templates.
"""

from samcli.commands.exceptions import UserException
from samcli.lib.init import generate_project
from samcli.lib.init.exceptions import InitErrorException


def do_generate(
    location,
    package_type,
    runtime,
    dependency_manager,
    output_dir,
    name,
    no_input,
    extra_context,
    tracing,
    application_insights,
):
    try:
        generate_project(
            location,
            package_type,
            runtime,
            dependency_manager,
            output_dir,
            name,
            no_input,
            extra_context,
            tracing,
            application_insights,
        )
    except InitErrorException as e:
        raise UserException(str(e), wrapped_from=e.__class__.__name__) from e



"""
Represents a pipeline OIDC provider
"""
from dataclasses import dataclass
from typing import Optional


@dataclass
class OidcConfig:
    oidc_provider_url: Optional[str]
    oidc_client_id: Optional[str]
    oidc_provider: Optional[str]

    def get_oidc_parameters(self) -> dict:
        return {
            "oidc-provider-url": self.oidc_provider_url,
            "oidc-client-id": self.oidc_client_id,
            "oidc-provider": self.oidc_provider,
        }

    def update_values(
        self, oidc_provider: Optional[str], oidc_provider_url: Optional[str], oidc_client_id: Optional[str]
    ) -> None:
        self.oidc_provider = oidc_provider if oidc_provider else self.oidc_provider
        self.oidc_provider_url = oidc_provider_url if oidc_provider_url else self.oidc_provider_url
        self.oidc_client_id = oidc_client_id if oidc_client_id else self.oidc_client_id


@dataclass
class GitHubOidcConfig:
    github_org: Optional[str]
    github_repo: Optional[str]
    deployment_branch: Optional[str]

    def get_oidc_parameters(self) -> dict:
        return {
            "github-org": self.github_org,
            "github-repo": self.github_repo,
            "deployment-branch": self.deployment_branch,
        }

    def update_values(
        self, github_org: Optional[str], github_repo: Optional[str], deployment_branch: Optional[str]
    ) -> None:
        self.github_org = github_org if github_org else self.github_org
        self.github_repo = github_repo if github_repo else self.github_repo
        self.deployment_branch = deployment_branch if deployment_branch else self.deployment_branch


@dataclass
class GitLabOidcConfig:
    gitlab_group: Optional[str]
    gitlab_project: Optional[str]
    deployment_branch: Optional[str]

    def get_oidc_parameters(self) -> dict:
        return {
            "gitlab-group": self.gitlab_group,
            "gitlab-project": self.gitlab_project,
            "deployment-branch": self.deployment_branch,
        }

    def update_values(
        self, gitlab_group: Optional[str], gitlab_project: Optional[str], deployment_branch: Optional[str]
    ) -> None:
        self.gitlab_group = gitlab_group if gitlab_group else self.gitlab_group
        self.gitlab_project = gitlab_project if gitlab_project else self.gitlab_project
        self.deployment_branch = deployment_branch if deployment_branch else self.deployment_branch


@dataclass
class BitbucketOidcConfig:
    bitbucket_repo_uuid: Optional[str]

    def get_oidc_parameters(self) -> dict:
        return {
            "bitbucket-repo-uuid": self.bitbucket_repo_uuid,
        }

    def update_values(self, bitbucket_repo_uuid: Optional[str]) -> None:
        self.bitbucket_repo_uuid = bitbucket_repo_uuid if bitbucket_repo_uuid else self.bitbucket_repo_uuid



"""
Keeps XRay event definitions
"""
import json
import operator
from typing import List, Optional

from samcli.lib.observability.observability_info_puller import ObservabilityEvent
from samcli.lib.utils.hash import str_checksum

start_time_getter = operator.attrgetter("start_time")


class XRayTraceEvent(ObservabilityEvent[dict]):
    """
    Represents a result of each XRay trace event, which is returned by boto3 client by calling 'batch_get_traces'
    See XRayTracePuller
    """

    def __init__(self, event: dict, revision: Optional[int] = None):
        super().__init__(event, 0)
        self.id = event.get("Id", "")
        # A revision number will be passed to link with the event
        # The same x-ray event will differ in information on different revisions
        self.revision = revision
        self.duration = event.get("Duration", 0.0)
        self.message = json.dumps(event)
        self.segments: List[XRayTraceSegment] = []

        self._construct_segments(event)
        if self.segments:
            self.timestamp = self.segments[0].start_time

    def _construct_segments(self, event_dict):
        """
        Each event is represented by segment, and it is like a Tree model (each segment also have subsegments).
        """
        raw_segments = event_dict.get("Segments", [])
        for raw_segment in raw_segments:
            segment_document = raw_segment.get("Document", "{}")
            self.segments.append(XRayTraceSegment(json.loads(segment_document)))
        self.segments.sort(key=start_time_getter)

    def get_latest_event_time(self):
        """
        Returns the latest event time for this specific XRayTraceEvent by calling get_latest_event_time for each segment
        """
        latest_event_time = 0
        for segment in self.segments:
            segment_latest_event_time = segment.get_latest_event_time()
            if segment_latest_event_time > latest_event_time:
                latest_event_time = segment_latest_event_time

        return latest_event_time


class XRayTraceSegment:
    """
    Represents each segment information for a XRayTraceEvent
    """

    def __init__(self, document: dict):
        self.id = document.get("Id", "")
        self.document = document
        self.name = document.get("name", "")
        self.start_time = document.get("start_time", 0)
        self.end_time = document.get("end_time", 0)
        self.http_status = document.get("http", {}).get("response", {}).get("status", None)
        self.sub_segments: List[XRayTraceSegment] = []

        sub_segments = document.get("subsegments", [])
        for sub_segment in sub_segments:
            self.sub_segments.append(XRayTraceSegment(sub_segment))
        self.sub_segments.sort(key=start_time_getter)

    def get_duration(self):
        return self.end_time - self.start_time

    def get_latest_event_time(self):
        """
        Gets the latest event time by comparing all timestamps (end_time) from current segment and all sub-segments
        """
        latest_event_time = self.end_time
        for sub_segment in self.sub_segments:
            sub_segment_latest_time = sub_segment.get_latest_event_time()
            if sub_segment_latest_time > latest_event_time:
                latest_event_time = sub_segment_latest_time

        return latest_event_time


class XRayServiceGraphEvent(ObservabilityEvent[dict]):
    """
    Represents a result of each XRay service graph event, which is returned by boto3 client by calling
    'get_service_graph' See XRayServiceGraphPuller
    """

    def __init__(self, event: dict):
        self.services: List[XRayGraphServiceInfo] = []
        self.message = str(event)
        self._construct_service(event)
        self.start_time = event.get("StartTime", None)
        self.end_time = event.get("EndTime", None)
        super().__init__(event, 0)

    def _construct_service(self, event_dict):
        services = event_dict.get("Services", [])
        for service in services:
            self.services.append(XRayGraphServiceInfo(service))

    def get_hash(self):
        """
        get the hash of the containing services
        """
        services = self.event.get("Services", [])
        return str_checksum(str(services))


class XRayGraphServiceInfo:
    """
    Represents each services information for a XRayServiceGraphEvent
    """

    def __init__(self, service: dict):
        self.id = service.get("ReferenceId", "")
        self.document = service
        self.name = service.get("Name", "")
        self.is_root = service.get("Root", False)
        self.type = service.get("Type")
        self.edge_ids: List[int] = []
        self.ok_count = 0
        self.error_count = 0
        self.fault_count = 0
        self.total_count = 0
        self.response_time = 0
        self._construct_edge_ids(service.get("Edges", []))
        self._set_summary_statistics(service.get("SummaryStatistics", None))

    def _construct_edge_ids(self, edges):
        """
        covert the edges information to a list of edge reference ids
        """
        edge_ids: List[int] = []
        for edge in edges:
            edge_ids.append(edge.get("ReferenceId", -1))
        self.edge_ids = edge_ids

    def _set_summary_statistics(self, summary_statistics):
        """
        get some useful information from summary statistics
        """
        if not summary_statistics:
            return
        self.ok_count = summary_statistics.get("OkCount", 0)
        error_statistics = summary_statistics.get("ErrorStatistics", None)
        if error_statistics:
            self.error_count = error_statistics.get("TotalCount", 0)
        fault_statistics = summary_statistics.get("FaultStatistics", None)
        if fault_statistics:
            self.fault_count = fault_statistics.get("TotalCount", 0)
        self.total_count = summary_statistics.get("TotalCount", 0)
        self.response_time = summary_statistics.get("TotalResponseTime", 0)



"""
Implementation of the stack output to table mapper
"""
from collections import OrderedDict
from typing import Any, Dict

from samcli.lib.list.list_interfaces import Mapper


class StackOutputToTableMapper(Mapper):
    """
    Mapper class for mapping stack-outputs data for table output
    """

    def map(self, data: list) -> Dict[Any, Any]:
        """
        Maps data to the format needed for consumption by the table consumer

        Parameters
        ----------
        data: list
            List of dictionaries containing the entries of the stack outputs data

        Returns
        -------
        table_data: Dict[Any, Any]
            Dictionary containing the information and data needed for the table consumer
            to output the data in table format
        """
        entry_list = []
        for stack_output in data:
            entry_list.append(
                [
                    stack_output.get("OutputKey", "-"),
                    stack_output.get("OutputValue", "-"),
                    stack_output.get("Description", "-"),
                ]
            )
        table_data = {
            "format_string": "{OutputKey:<{0}} {OutputValue:<{1}} {Description:<{2}}",
            "format_args": OrderedDict(
                {"OutputKey": "OutputKey", "OutputValue": "OutputValue", "Description": "Description"}
            ),
            "table_name": "Stack Outputs",
            "data": entry_list,
        }
        return table_data



"""
Custom exceptions raised by this local library
"""
from samcli.commands.exceptions import UserException


class NoApisDefined(Exception):
    """
    Raised when there are no APIs defined in the template
    """


class OverridesNotWellDefinedError(Exception):
    """
    Raised when the overrides file is invalid
    """


class NoPrivilegeException(Exception):
    """
    Process does not have the required privilege to complete the action
    """


class InvalidIntermediateImageError(Exception):
    """
    Raised when there is no valid intermediate image to build on top of
    for Image based PackageTypes.
    """


class UnsupportedRuntimeArchitectureError(UserException):
    """
    Raised when runtime does not support architecture
    """


class UnsupportedInlineCodeError(UserException):
    """
    Raised when inline code is used for sam local commands
    """



"""
Read and parse CLI args for the Logs Command and setup the context for running the command
"""

import logging
from typing import Any, Dict, List, Optional, Set

from samcli.commands.exceptions import UserException
from samcli.lib.utils.boto_utils import BotoProviderType
from samcli.lib.utils.cloudformation import CloudFormationResourceSummary, get_resource_summaries
from samcli.lib.utils.resources import (
    AWS_APIGATEWAY_RESTAPI,
    AWS_APIGATEWAY_V2_API,
    AWS_LAMBDA_FUNCTION,
    AWS_STEPFUNCTIONS_STATEMACHINE,
)
from samcli.lib.utils.time import parse_date, to_utc

LOG = logging.getLogger(__name__)


class InvalidTimestampError(UserException):
    """
    Used to indicate that given date time string is an invalid timestamp
    """


class TimeParseError(UserException):
    """
    Used to throw if parsing of the given time string or UTC conversion is failed
    """


def parse_time(time_str: str, property_name: str):
    """
    Parse the time from the given string, convert to UTC, and return the datetime object

    Parameters
    ----------
    time_str : str
        The time to parse

    property_name : str
        Name of the property where this time came from. Used in the exception raised if time is not parseable

    Returns
    -------
    datetime.datetime
        Parsed datetime object

    Raises
    ------
    InvalidTimestampError
        If the string cannot be parsed as a timestamp
    """
    try:
        if not time_str:
            return None

        parsed = parse_date(time_str)
        if not parsed:
            raise InvalidTimestampError(f"Unable to parse the time provided by '{property_name}'")

        return to_utc(parsed)
    except InvalidTimestampError as ex:
        raise ex
    except Exception as ex:
        LOG.error("Failed to parse given time information %s", time_str, exc_info=ex)
        raise TimeParseError(f"Unable to parse the time information '{property_name}': '{time_str}'") from ex


class ResourcePhysicalIdResolver:
    """
    Wrapper class that is used to extract information about resources which we can tail their logs for given stack
    """

    # list of resource types that is supported right now for pulling their logs
    DEFAULT_SUPPORTED_RESOURCES: Set[str] = {
        AWS_LAMBDA_FUNCTION,
        AWS_APIGATEWAY_RESTAPI,
        AWS_APIGATEWAY_V2_API,
        AWS_STEPFUNCTIONS_STATEMACHINE,
    }

    def __init__(
        self,
        boto_resource_provider: BotoProviderType,
        boto_client_provider: BotoProviderType,
        stack_name: str,
        resource_names: Optional[List[str]] = None,
        supported_resource_types: Optional[Set[str]] = None,
    ):
        self._boto_resource_provider = boto_resource_provider
        self._boto_client_provider = boto_client_provider
        self._stack_name = stack_name
        if resource_names is None:
            resource_names = []
        if supported_resource_types is None:
            supported_resource_types = ResourcePhysicalIdResolver.DEFAULT_SUPPORTED_RESOURCES
        self._supported_resource_types: Set[str] = supported_resource_types
        self._resource_names = set(resource_names)

    def get_resource_information(self, fetch_all_when_no_resource_name_given: bool = True) -> List[Any]:
        """
        Returns the list of resource information for the given stack.

        Parameters
        ----------
        fetch_all_when_no_resource_name_given : bool
            When given, it will fetch all resources if no specific resource name is provided, default value is True

        Returns
        -------
        List[StackResourceSummary]
            List of resource information, which will be used to fetch the logs
        """
        if self._resource_names:
            return self._fetch_resources_from_stack(self._resource_names)
        if fetch_all_when_no_resource_name_given:
            return self._fetch_resources_from_stack()
        return []

    def _fetch_resources_from_stack(
        self, selected_resource_names: Optional[Set[str]] = None
    ) -> List[CloudFormationResourceSummary]:
        """
        Returns list of all resources from given stack name
        If any resource is not supported, it will discard them

        Parameters
        ----------
        selected_resource_names : Optional[Set[str]]
            An optional set of string parameter, which will filter resource names. If none is given, it will be
            equal to all resource names in stack, which means there won't be any filtering by resource name.

        Returns
        -------
        List[CloudFormationResourceSummary]
            List of resource information, which will be used to fetch the logs
        """
        LOG.debug("Getting logical id of the all resources for stack '%s'", self._stack_name)
        stack_resources = get_resource_summaries(
            self._boto_resource_provider,
            self._boto_client_provider,
            self._stack_name,
            ResourcePhysicalIdResolver.DEFAULT_SUPPORTED_RESOURCES,
        )

        if selected_resource_names:
            return self._get_selected_resources(stack_resources, selected_resource_names)
        return list(stack_resources.values())

    @staticmethod
    def _get_selected_resources(
        resource_summaries: Dict[str, CloudFormationResourceSummary],
        selected_resource_names: Set[str],
    ) -> List[CloudFormationResourceSummary]:
        """
        Returns list of resources which matches with selected_resource_names.
        selected_resource_names can be;
        - resource name like HelloWorldFunction
        - or it could be pointing to a resource in nested stack like NestedApp/HelloWorldFunction

        Parameters
        ----------
        resource_summaries : Dict[str, CloudFormationResourceSummary]
            Dictionary of resource key and CloudformationResourceSummary which was returned from given stack
        selected_resource_names : Set[str]
            List of resource name definitions that will be used to filter the results

        Returns
        ------
        List[CloudFormationResourceSummary]
            Filtered list of CloudFormationResourceSummary's
        """
        resources = []
        for selected_resource_name in selected_resource_names:
            selected_resource = resource_summaries.get(selected_resource_name)
            if selected_resource:
                resources.append(selected_resource)
            else:
                LOG.warning(
                    "Resource name (%s) does not exist. Available resource names: %s",
                    selected_resource_name,
                    ", ".join(resource_summaries.keys()),
                )
        return resources



"""
Client for uploading packaged artifacts to ecr
"""
import base64
import io
import logging
from typing import Dict

import botocore
import click
import docker
from docker.errors import APIError, BuildError

from samcli.commands.package.exceptions import (
    DeleteArtifactFailedError,
    DockerLoginFailedError,
    DockerPushFailedError,
    ECRAuthorizationError,
)
from samcli.lib.docker.log_streamer import LogStreamer, LogStreamError
from samcli.lib.package.image_utils import tag_translation
from samcli.lib.utils.osutils import stderr
from samcli.lib.utils.stream_writer import StreamWriter

LOG = logging.getLogger(__name__)

ECR_USERNAME = "AWS"


class ECRUploader:
    """
    Class to upload Images to ECR.
    """

    def __init__(
        self, docker_client, ecr_client, ecr_repo, ecr_repo_multi, no_progressbar=False, tag="latest", stream=stderr()
    ):
        self.docker_client = docker_client if docker_client else docker.from_env()
        self.ecr_client = ecr_client
        self.ecr_repo = ecr_repo
        self.ecr_repo_multi = ecr_repo_multi
        self.tag = tag
        self.auth_config = {}
        self.no_progressbar = no_progressbar
        self.stream = StreamWriter(stream=stream, auto_flush=True)
        self.log_streamer = LogStreamer(stream=self.stream)
        self.login_session_active = False

    def login(self):
        """
        Logs into the supplied ECR with credentials.
        """
        try:
            token = self.ecr_client.get_authorization_token()
        except botocore.exceptions.ClientError as ex:
            raise ECRAuthorizationError(msg=ex.response["Error"]["Message"]) from ex

        username, password = base64.b64decode(token["authorizationData"][0]["authorizationToken"]).decode().split(":")
        registry = token["authorizationData"][0]["proxyEndpoint"]

        try:
            self.docker_client.login(username=ECR_USERNAME, password=password, registry=registry)
        except APIError as ex:
            raise DockerLoginFailedError(msg=str(ex)) from ex
        self.auth_config = {"username": username, "password": password}

    def upload(self, image, resource_name):
        """
        Uploads given local image to ECR.
        :param image: locally tagged docker image that would be uploaded to ECR.
        :param resource_name: logical ID of the resource to be uploaded to ECR.
        :return: remote ECR image path that has been uploaded.
        """
        if not self.login_session_active:
            self.login()
            self.login_session_active = True
        try:
            docker_img = self.docker_client.images.get(image)

            _tag = tag_translation(image, docker_image_id=docker_img.id, gen_tag=self.tag)
            repository = (
                self.ecr_repo
                if not self.ecr_repo_multi or not isinstance(self.ecr_repo_multi, dict)
                else self.ecr_repo_multi.get(resource_name)
            )

            docker_img.tag(repository=repository, tag=_tag)
            push_logs = self.docker_client.api.push(
                repository=repository, tag=_tag, auth_config=self.auth_config, stream=True, decode=True
            )
            if not self.no_progressbar:
                self.log_streamer.stream_progress(push_logs)
            else:
                # we need to wait till the image got pushed to ecr, without this workaround sam sync for template
                # contains image always fail, because the provided ecr uri is not exist.
                _log_streamer = LogStreamer(stream=StreamWriter(stream=io.BytesIO(), auto_flush=True))
                _log_streamer.stream_progress(push_logs)

        except (BuildError, APIError, LogStreamError) as ex:
            raise DockerPushFailedError(msg=str(ex)) from ex

        return f"{repository}:{_tag}"

    def delete_artifact(self, image_uri: str, resource_id: str, property_name: str):
        """
        Delete the given ECR image by extracting the repository and image_tag from
        image_uri

        :param image_uri: image_uri of the image to be deleted
        :param resource_id: id of the resource for which the image is deleted
        :param property_name: provided property_name for the resource
        """
        try:
            repo_image_tag = self.parse_image_url(image_uri=image_uri)
            repository = repo_image_tag["repository"]
            image_tag = repo_image_tag["image_tag"]
            resp = self.ecr_client.batch_delete_image(
                repositoryName=repository,
                imageIds=[
                    {"imageTag": image_tag},
                ],
            )
            if resp["failures"]:
                # Image not found
                image_details = resp["failures"][0]
                if image_details["failureCode"] == "ImageNotFound":
                    LOG.debug(
                        "Could not delete image for %s parameter of %s resource as it does not exist. \n",
                        property_name,
                        resource_id,
                    )
                    click.echo(f"\t- Could not find image with tag {image_tag} in repository {repository}")
                else:
                    LOG.debug(
                        "Could not delete the image for the resource %s. FailureCode: %s, FailureReason: %s",
                        property_name,
                        image_details["failureCode"],
                        image_details["failureReason"],
                    )
                    click.echo(f"\t- Could not delete image with tag {image_tag} in repository {repository}")
            else:
                LOG.debug("Deleting ECR image with tag %s", image_tag)
                click.echo(f"\t- Deleting ECR image {image_tag} in repository {repository}")

        except botocore.exceptions.ClientError as ex:
            # Handle Client errors such as RepositoryNotFoundException or InvalidParameterException
            if "RepositoryNotFoundException" not in str(ex):
                LOG.debug("DeleteArtifactFailedError Exception : %s", str(ex))
                raise DeleteArtifactFailedError(resource_id=resource_id, property_name=property_name, ex=ex) from ex
            LOG.debug("RepositoryNotFoundException : %s", str(ex))

    def delete_ecr_repository(self, physical_id: str):
        """
        Delete ECR repository using the physical_id

        :param: physical_id of the repository to be deleted
        """
        try:
            click.echo(f"\t- Deleting ECR repository {physical_id}")
            self.ecr_client.delete_repository(repositoryName=physical_id, force=True)
        except self.ecr_client.exceptions.RepositoryNotFoundException:
            # If the repository is empty, cloudformation automatically deletes
            # the repository when cf_client.delete_stack is called.
            LOG.debug("Could not find repository %s", physical_id)

    @staticmethod
    def parse_image_url(image_uri: str) -> Dict:
        result = {}
        registry_repo_tag = image_uri.split("/", 1)
        repo_colon_image_tag = None
        if len(registry_repo_tag) == 1:
            # If there is no registry specified, e.g. repo:tag
            repo_colon_image_tag = registry_repo_tag[0]
        else:
            # Registry present, e.g. registry/repo:tag
            repo_colon_image_tag = registry_repo_tag[1]
        repo_image_tag_split = repo_colon_image_tag.split(":")

        # If no tag is specified, use latest
        result["repository"] = repo_image_tag_split[0]
        result["image_tag"] = repo_image_tag_split[1] if len(repo_image_tag_split) > 1 else "latest"

        return result



"""
CLI command for "package" command
"""
import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.cdk_support_decorators import unsupported_command_cdk
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.options import (
    force_upload_option,
    image_repositories_option,
    image_repository_option,
    kms_key_id_option,
    metadata_option,
    no_progressbar_option,
    resolve_s3_option,
    s3_bucket_option,
    s3_prefix_option,
    signing_profiles_option,
    template_click_option,
    use_json_option,
)
from samcli.lib.bootstrap.bootstrap import manage_stack
from samcli.lib.cli_validation.image_repository_validation import image_repository_validation
from samcli.lib.telemetry.metric import track_command, track_template_warnings
from samcli.lib.utils.resources import resources_generator
from samcli.lib.utils.version_checker import check_newer_version
from samcli.lib.warnings.sam_cli_warning import CodeDeployConditionWarning, CodeDeployWarning

SHORT_HELP = "Package an AWS SAM application."


def resources_and_properties_help_string():
    """
    Total list of resources and their property locations that are supported for `sam package`
    :return: str
    """
    return "".join(
        f"\nResource : {resource} | Location : {location}\n".format(resource=resource, location=location)
        for resource, location in resources_generator()
    )


HELP_TEXT = (
    """The SAM package command creates and uploads artifacts based on the package type of a given resource.
It uploads local images to ECR for `Image` package types.
It creates zip of your code and dependencies and uploads it to S3 for other package types.
The command returns a copy of your template, replacing references to local artifacts
with the AWS location where the command uploaded the artifacts.

The following resources and their property locations are supported.
"""
    + resources_and_properties_help_string()
)


@click.command("package", short_help=SHORT_HELP, help=HELP_TEXT, context_settings=dict(max_content_width=120))
@configuration_option(provider=TomlProvider(section="parameters"))
@template_click_option(include_build=True)
@click.option(
    "--output-template-file",
    required=False,
    type=click.Path(),
    help="The path to the file where the command "
    "writes the output AWS CloudFormation template. If you don't specify a "
    "path, the command writes the template to the standard output.",
)
@s3_bucket_option
@image_repository_option
@image_repositories_option
@s3_prefix_option
@kms_key_id_option
@use_json_option
@force_upload_option
@resolve_s3_option
@metadata_option
@signing_profiles_option
@no_progressbar_option
@common_options
@aws_creds_options
@image_repository_validation
@pass_context
@track_command
@check_newer_version
@track_template_warnings([CodeDeployWarning.__name__, CodeDeployConditionWarning.__name__])
@print_cmdline_args
@unsupported_command_cdk(alternative_command="cdk deploy")
@command_exception_handler
def cli(
    ctx,
    template_file,
    s3_bucket,
    image_repository,
    image_repositories,
    s3_prefix,
    kms_key_id,
    output_template_file,
    use_json,
    force_upload,
    no_progressbar,
    metadata,
    signing_profiles,
    resolve_s3,
    config_file,
    config_env,
):
    """
    `sam package` command entry point
    """
    # All logic must be implemented in the ``do_cli`` method. This helps with easy unit testing

    do_cli(
        template_file,
        s3_bucket,
        image_repository,
        image_repositories,
        s3_prefix,
        kms_key_id,
        output_template_file,
        use_json,
        force_upload,
        no_progressbar,
        metadata,
        signing_profiles,
        ctx.region,
        ctx.profile,
        resolve_s3,
    )  # pragma: no cover


def do_cli(
    template_file,
    s3_bucket,
    image_repository,
    image_repositories,
    s3_prefix,
    kms_key_id,
    output_template_file,
    use_json,
    force_upload,
    no_progressbar,
    metadata,
    signing_profiles,
    region,
    profile,
    resolve_s3,
):
    """
    Implementation of the ``cli`` method
    """

    from samcli.commands.package.package_context import PackageContext

    if resolve_s3:
        s3_bucket = manage_stack(profile=profile, region=region)
        click.echo(f"\n\t\tManaged S3 bucket: {s3_bucket}")
        click.echo("\t\tA different default S3 bucket can be set in samconfig.toml")
        click.echo("\t\tOr by specifying --s3-bucket explicitly.")

    with PackageContext(
        template_file=template_file,
        s3_bucket=s3_bucket,
        image_repository=image_repository,
        image_repositories=image_repositories,
        s3_prefix=s3_prefix,
        kms_key_id=kms_key_id,
        output_template_file=output_template_file,
        use_json=use_json,
        force_upload=force_upload,
        no_progressbar=no_progressbar,
        metadata=metadata,
        region=region,
        profile=profile,
        signing_profiles=signing_profiles,
    ) as package_context:
        package_context.run()



"""Build exceptions"""

from samcli.commands.exceptions import UserException


class InvalidBuildDirException(UserException):
    """
    Value provided to --build-dir is invalid
    """


class MissingBuildMethodException(UserException):
    """
    Exception to be thrown when a layer is tried to build without BuildMethod
    """



"""
Class containing error conditions that are exposed to the user.
"""

from samcli.commands.exceptions import UserException


class InvokeContextException(UserException):
    """
    Something went wrong invoking the function.
    """


class InvalidSamTemplateException(UserException):
    """
    The template provided was invalid and not able to transform into a Standard CloudFormation Template
    """


class SamTemplateNotFoundException(UserException):
    """
    The SAM Template provided could not be found
    """


class DebugContextException(UserException):
    """
    Something went wrong when creating the DebugContext
    """


class ImageBuildException(UserException):
    """
    Image failed to build
    """


class CredentialsRequired(UserException):
    """
    Credentials were not given when Required
    """


class ResourceNotFound(UserException):
    """
    The Resource requested was not found
    """


class InvalidLayerVersionArn(UserException):
    """
    The LayerVersion Arn given in the template is Invalid
    """


class UnsupportedIntrinsic(UserException):
    """
    Value from a template has an Intrinsic that is unsupported
    """


class NotAvailableInRegion(UserException):
    """
    Calling service not available (launched) in specified region
    """


class InvalidFunctionPropertyType(UserException):
    """
    Function property given in the template is Invalid
    """



"""
CLI command for "pipeline bootstrap", which sets up the require pipeline infrastructure resources
"""
import os
from textwrap import dedent
from typing import Any, Dict, List, Optional

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.click_mutex import ClickMutex
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands.pipeline.bootstrap.oidc_config import (
    BitbucketOidcConfig,
    GitHubOidcConfig,
    GitLabOidcConfig,
    OidcConfig,
)
from samcli.commands.pipeline.bootstrap.pipeline_oidc_provider import PipelineOidcProvider
from samcli.lib.config.samconfig import SamConfig
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

from .guided_context import BITBUCKET, GITHUB_ACTIONS, GITLAB, IAM, OPEN_ID_CONNECT

SHORT_HELP = "Generates the required AWS resources to connect your CI/CD system."

HELP_TEXT = """
This command generates the required AWS infrastructure resources to connect to your CI/CD system.
This step must be run for each deployment stage in your pipeline, prior to running the sam pipline init command.
"""

PIPELINE_CONFIG_DIR = os.path.join(".aws-sam", "pipeline")
PIPELINE_CONFIG_FILENAME = "pipelineconfig.toml"
PERMISSIONS_PROVIDERS = [OPEN_ID_CONNECT, IAM]
OPENID_CONNECT = "OpenID Connect (OIDC)"


@click.command("bootstrap", short_help=SHORT_HELP, help=HELP_TEXT, context_settings=dict(max_content_width=120))
@configuration_option(provider=TomlProvider(section="parameters"))
@click.option(
    "--interactive/--no-interactive",
    is_flag=True,
    default=True,
    help="Disable interactive prompting for bootstrap parameters, and fail if any required arguments are missing.",
)
@click.option(
    "--stage",
    help="The name of the corresponding deployment stage. "
    "It is used as a suffix for the created AWS infrastructure resources.",
    required=False,
)
@click.option(
    "--pipeline-user",
    help="The Amazon Resource Name (ARN) of the IAM user having its access key ID and secret access key "
    "shared with the CI/CD system. It is used to grant this IAM user permission to access the "
    "corresponding AWS account. If not provided, the command will create one along with the access "
    "key ID and secret access key credentials.",
    required=False,
)
@click.option(
    "--pipeline-execution-role",
    help="The ARN of the IAM role to be assumed by the pipeline user to operate on this stage. "
    "Provide it only if you want to use your own role, otherwise this command will create one.",
    required=False,
)
@click.option(
    "--cloudformation-execution-role",
    help="The ARN of the IAM role to be assumed by the AWS CloudFormation service while deploying the "
    "application's stack. Provide only if you want to use your own role, otherwise the command will create one.",
    required=False,
)
@click.option(
    "--bucket",
    help="The ARN of the Amazon S3 bucket to hold the AWS SAM artifacts.",
    required=False,
)
@click.option(
    "--create-image-repository/--no-create-image-repository",
    is_flag=True,
    default=False,
    help="If set to true and no ECR image repository is provided, this command will create an ECR image repository "
    "to hold the container images of Lambda functions having an Image package type.",
)
@click.option(
    "--image-repository",
    help="The ARN of an Amazon ECR image repository to hold the container images of Lambda functions or "
    "layers that have a package type of Image. If provided, the --create-image-repository options is ignored. "
    "If not provided and --create-image-repository is specified, the command will create one.",
    required=False,
)
@click.option(
    "--confirm-changeset/--no-confirm-changeset",
    default=True,
    is_flag=True,
    help="Prompt to confirm if the resources are to be deployed.",
)
@click.option(
    "--permissions-provider",
    default=IAM,
    required=False,
    type=click.Choice(PERMISSIONS_PROVIDERS),
    help="Choose a permissions provider to assume the pipeline execution role. Default is to use an IAM User.",
)
@click.option(
    "--oidc-provider-url",
    help="The URL of the OIDC provider.",
    required=False,
)
@click.option("--oidc-client-id", help="The client ID configured to use with the OIDC provider.", required=False)
@click.option(
    "--github-org",
    help="The GitHub organization that the repository belongs to. "
    "If there is no organization enter the Username of the repository owner instead "
    "Only used if using GitHub Actions OIDC for user permissions",
    required=False,
    cls=ClickMutex,
    incompatible_params=["bitbucket_repo_uuid", "gitlab_group", "gitlab_project"],
)
@click.option(
    "--github-repo",
    help="The name of the GitHub Repository that deployments will occur from. "
    "Only used if using GitHub Actions OIDC for permissions",
    required=False,
    cls=ClickMutex,
    incompatible_params=["bitbucket_repo_uuid", "gitlab_group", "gitlab_project"],
)
@click.option(
    "--deployment-branch",
    help="The name of the branch that deployments will occur from. "
    "Only used if using GitHub Actions OIDC for permissions",
    required=False,
)
@click.option(
    "--oidc-provider",
    help="The name of the CI/CD system that will be used for OIDC permissions "
    "Currently supported CI/CD systems are : GitLab, GitHub and Bitbucket",
    type=click.Choice([GITHUB_ACTIONS, GITLAB, BITBUCKET]),
    required=False,
    cls=ClickMutex,
    required_param_lists=[
        ["gitlab_group", "gitlab_project"],
        ["github_org", "github_repo"],
        ["bitbucket_repo_uuid"],
    ],
)
@click.option(
    "--gitlab-group",
    help="The GitLab group that the repository belongs to. Only used if using GitLab OIDC for permissions",
    required=False,
    cls=ClickMutex,
    incompatible_params=["bitbucket_repo_uuid", "github_org", "github_repo"],
)
@click.option(
    "--gitlab-project",
    help="The GitLab project name. Only used if using GitLab OIDC for permissions",
    required=False,
    cls=ClickMutex,
    incompatible_params=["bitbucket_repo_uuid", "github_org", "github_repo"],
)
@click.option(
    "--bitbucket-repo-uuid",
    help="The UUID of the Bitbucket repository. Only used if using Bitbucket OIDC for permissions. "
    "Found at https://bitbucket.org/<WORKSPACE>/<REPOSITORY>/admin/addon/admin/pipelines/openid-connect",
    required=False,
    cls=ClickMutex,
    incompatible_params=["gitlab_group", "gitlab_project", "github_org", "github_repo"],
)
@click.option(
    "--cicd-provider",
    help="The CICD platform for the SAM Pipeline",
    required=False,
)
@common_options
@aws_creds_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(
    ctx: Any,
    interactive: bool,
    stage: Optional[str],
    pipeline_user: Optional[str],
    pipeline_execution_role: Optional[str],
    cloudformation_execution_role: Optional[str],
    bucket: Optional[str],
    create_image_repository: bool,
    image_repository: Optional[str],
    confirm_changeset: bool,
    config_file: Optional[str],
    config_env: Optional[str],
    permissions_provider: Optional[str],
    oidc_provider_url: Optional[str],
    oidc_client_id: Optional[str],
    github_org: Optional[str],
    github_repo: Optional[str],
    deployment_branch: Optional[str],
    oidc_provider: Optional[str],
    gitlab_group: Optional[str],
    gitlab_project: Optional[str],
    bitbucket_repo_uuid: Optional[str],
    cicd_provider: Optional[str],
) -> None:
    """
    `sam pipeline bootstrap` command entry point
    """
    do_cli(
        region=ctx.region,
        profile=ctx.profile,
        interactive=interactive,
        stage_configuration_name=stage,
        pipeline_user_arn=pipeline_user,
        pipeline_execution_role_arn=pipeline_execution_role,
        cloudformation_execution_role_arn=cloudformation_execution_role,
        artifacts_bucket_arn=bucket,
        create_image_repository=create_image_repository,
        image_repository_arn=image_repository,
        confirm_changeset=confirm_changeset,
        config_file=config_env,
        config_env=config_file,
        permissions_provider=permissions_provider,
        oidc_provider_url=oidc_provider_url,
        oidc_client_id=oidc_client_id,
        github_org=github_org,
        github_repo=github_repo,
        deployment_branch=deployment_branch,
        oidc_provider=oidc_provider,
        gitlab_group=gitlab_group,
        gitlab_project=gitlab_project,
        bitbucket_repo_uuid=bitbucket_repo_uuid,
        cicd_provider=cicd_provider,
    )  # pragma: no cover


def do_cli(
    region: Optional[str],
    profile: Optional[str],
    interactive: bool,
    stage_configuration_name: Optional[str],
    pipeline_user_arn: Optional[str],
    pipeline_execution_role_arn: Optional[str],
    cloudformation_execution_role_arn: Optional[str],
    artifacts_bucket_arn: Optional[str],
    create_image_repository: bool,
    image_repository_arn: Optional[str],
    confirm_changeset: bool,
    config_file: Optional[str],
    config_env: Optional[str],
    permissions_provider: Optional[str],
    oidc_provider_url: Optional[str],
    oidc_client_id: Optional[str],
    github_org: Optional[str],
    github_repo: Optional[str],
    deployment_branch: Optional[str],
    oidc_provider: Optional[str],
    gitlab_group: Optional[str],
    gitlab_project: Optional[str],
    bitbucket_repo_uuid: Optional[str],
    cicd_provider: Optional[str],
    standalone: bool = True,
) -> None:
    """
    implementation of `sam pipeline bootstrap` command
    """
    from samcli.commands.pipeline.bootstrap.guided_context import GuidedContext
    from samcli.commands.pipeline.external_links import CONFIG_AWS_CRED_ON_CICD_URL
    from samcli.lib.pipeline.bootstrap.stage import (
        BITBUCKET_REPO_UUID,
        DEPLOYMENT_BRANCH,
        GITHUB_ORG,
        GITHUB_REPO,
        GITLAB_GROUP,
        GITLAB_PROJECT,
        OIDC_CLIENT_ID,
        OIDC_PROVIDER,
        OIDC_PROVIDER_URL,
        OIDC_SUPPORTED_PROVIDER,
        PERMISSIONS_PROVIDER,
        Stage,
    )
    from samcli.lib.utils.colors import Colored

    config_parameters = _load_config_values()
    if not pipeline_user_arn and not permissions_provider == OPEN_ID_CONNECT:
        pipeline_user_arn = config_parameters.get("pipeline_user")

    enable_oidc_option = False
    if not cicd_provider or cicd_provider in OIDC_SUPPORTED_PROVIDER:
        enable_oidc_option = True
        oidc_provider = cicd_provider

    oidc_config = OidcConfig(
        oidc_client_id=oidc_client_id, oidc_provider=oidc_provider, oidc_provider_url=oidc_provider_url
    )
    gitlab_config = GitLabOidcConfig(
        gitlab_group=gitlab_group, gitlab_project=gitlab_project, deployment_branch=deployment_branch
    )
    github_config = GitHubOidcConfig(
        github_org=github_org, github_repo=github_repo, deployment_branch=deployment_branch
    )
    bitbucket_config = BitbucketOidcConfig(bitbucket_repo_uuid=bitbucket_repo_uuid)
    if config_parameters:
        saved_provider = config_parameters.get(PERMISSIONS_PROVIDER)
        if saved_provider == OPENID_CONNECT:
            permissions_provider = OPEN_ID_CONNECT
            oidc_config.update_values(
                oidc_provider=config_parameters.get(OIDC_PROVIDER),
                oidc_provider_url=config_parameters.get(OIDC_PROVIDER_URL),
                oidc_client_id=config_parameters.get(OIDC_CLIENT_ID),
            )
            github_config.update_values(
                github_org=config_parameters.get(GITHUB_ORG),
                github_repo=config_parameters.get(GITHUB_REPO),
                deployment_branch=config_parameters.get(DEPLOYMENT_BRANCH),
            )
            gitlab_config.update_values(
                gitlab_group=config_parameters.get(GITLAB_GROUP),
                gitlab_project=config_parameters.get(GITLAB_PROJECT),
                deployment_branch=config_parameters.get(DEPLOYMENT_BRANCH),
            )
            bitbucket_config.update_values(bitbucket_repo_uuid=config_parameters.get(BITBUCKET_REPO_UUID))
        elif saved_provider == "AWS IAM":
            permissions_provider = IAM
    if interactive:
        if standalone:
            click.echo(
                dedent(
                    """\

                    sam pipeline bootstrap generates the required AWS infrastructure resources to connect
                    to your CI/CD system. This step must be run for each deployment stage in your pipeline,
                    prior to running the sam pipeline init command.

                    We will ask for [1] stage definition, [2] account details, and
                    [3] references to existing resources in order to bootstrap these pipeline resources.
                    """
                ),
            )

        guided_context = GuidedContext(
            profile=profile,
            stage_configuration_name=stage_configuration_name,
            pipeline_user_arn=pipeline_user_arn,
            pipeline_execution_role_arn=pipeline_execution_role_arn,
            cloudformation_execution_role_arn=cloudformation_execution_role_arn,
            artifacts_bucket_arn=artifacts_bucket_arn,
            create_image_repository=create_image_repository,
            image_repository_arn=image_repository_arn,
            region=region,
            permissions_provider=permissions_provider,
            oidc_config=oidc_config,
            github_config=github_config,
            gitlab_config=gitlab_config,
            bitbucket_config=bitbucket_config,
            enable_oidc_option=enable_oidc_option,
        )
        guided_context.run()
        stage_configuration_name = guided_context.stage_configuration_name
        pipeline_user_arn = guided_context.pipeline_user_arn
        pipeline_execution_role_arn = guided_context.pipeline_execution_role_arn
        cloudformation_execution_role_arn = guided_context.cloudformation_execution_role_arn
        artifacts_bucket_arn = guided_context.artifacts_bucket_arn
        create_image_repository = guided_context.create_image_repository
        image_repository_arn = guided_context.image_repository_arn
        region = guided_context.region
        profile = guided_context.profile
        permissions_provider = guided_context.permissions_provider

    subject_claim = None
    pipeline_oidc_provider: Optional[PipelineOidcProvider] = None

    if permissions_provider == OPEN_ID_CONNECT:
        pipeline_oidc_provider = _get_pipeline_oidc_provider(
            oidc_config=oidc_config,
            github_config=github_config,
            gitlab_config=gitlab_config,
            bitbucket_config=bitbucket_config,
        )
        subject_claim = pipeline_oidc_provider.get_subject_claim()

    if not stage_configuration_name:
        raise click.UsageError("Missing required parameter '--stage'")

    environment: Stage = Stage(
        name=stage_configuration_name,
        aws_profile=profile,
        aws_region=region,
        pipeline_user_arn=pipeline_user_arn,
        pipeline_execution_role_arn=pipeline_execution_role_arn,
        cloudformation_execution_role_arn=cloudformation_execution_role_arn,
        artifacts_bucket_arn=artifacts_bucket_arn,
        create_image_repository=create_image_repository,
        image_repository_arn=image_repository_arn,
        oidc_provider_url=oidc_config.oidc_provider_url,
        oidc_client_id=oidc_config.oidc_client_id,
        permissions_provider=permissions_provider,
        subject_claim=subject_claim,
        pipeline_oidc_provider=pipeline_oidc_provider,
    )

    bootstrapped: bool = environment.bootstrap(confirm_changeset=confirm_changeset)

    if bootstrapped:
        environment.print_resources_summary()

        environment.save_config_safe(
            config_dir=PIPELINE_CONFIG_DIR, filename=PIPELINE_CONFIG_FILENAME, cmd_names=_get_bootstrap_command_names()
        )

        click.secho(
            dedent(
                f"""\
                View the definition in {os.path.join(PIPELINE_CONFIG_DIR, PIPELINE_CONFIG_FILENAME)},
                run sam pipeline bootstrap to generate another set of resources, or proceed to
                sam pipeline init to create your pipeline configuration file.
                """
            )
        )

        if not environment.pipeline_user.is_user_provided and not environment.use_oidc_provider:
            click.secho(
                dedent(
                    f"""\
                    Before running {Colored().bold("sam pipeline init")}, we recommend first setting up AWS credentials
                    in your CI/CD account. Read more about how to do so with your provider in
                    {CONFIG_AWS_CRED_ON_CICD_URL}.
                    """
                )
            )


def _get_pipeline_oidc_provider(
    oidc_config: OidcConfig,
    github_config: GitHubOidcConfig,
    gitlab_config: GitLabOidcConfig,
    bitbucket_config: BitbucketOidcConfig,
) -> PipelineOidcProvider:
    from samcli.commands.pipeline.bootstrap.pipeline_oidc_provider import (
        BitbucketOidcProvider,
        GitHubOidcProvider,
        GitLabOidcProvider,
    )

    if oidc_config.oidc_provider == GITHUB_ACTIONS:
        github_oidc_params: dict = {
            GitHubOidcProvider.GITHUB_ORG_PARAMETER_NAME: github_config.github_org,
            GitHubOidcProvider.GITHUB_REPO_PARAMETER_NAME: github_config.github_repo,
            GitHubOidcProvider.DEPLOYMENT_BRANCH_PARAMETER_NAME: github_config.deployment_branch,
        }
        return GitHubOidcProvider(github_oidc_params, oidc_config.get_oidc_parameters())
    if oidc_config.oidc_provider == GITLAB:
        gitlab_oidc_params: dict = {
            GitLabOidcProvider.GITLAB_PROJECT_PARAMETER_NAME: gitlab_config.gitlab_project,
            GitLabOidcProvider.GITLAB_GROUP_PARAMETER_NAME: gitlab_config.gitlab_group,
            GitLabOidcProvider.DEPLOYMENT_BRANCH_PARAMETER_NAME: gitlab_config.deployment_branch,
        }
        return GitLabOidcProvider(gitlab_oidc_params, oidc_config.get_oidc_parameters())
    if oidc_config.oidc_provider == BITBUCKET:
        bitbucket_oidc_params: dict = {
            BitbucketOidcProvider.BITBUCKET_REPO_UUID_PARAMETER_NAME: bitbucket_config.bitbucket_repo_uuid
        }
        return BitbucketOidcProvider(bitbucket_oidc_params, oidc_config.get_oidc_parameters())
    raise click.UsageError("Missing required parameter '--oidc-provider'")


def _load_config_values() -> Dict[str, str]:
    samconfig: SamConfig = SamConfig(config_dir=PIPELINE_CONFIG_DIR, filename=PIPELINE_CONFIG_FILENAME)
    if not samconfig.exists():
        return {}
    config: Dict[str, str] = samconfig.get_all(cmd_names=_get_bootstrap_command_names(), section="parameters")
    return config


def _get_bootstrap_command_names() -> List[str]:
    return ["pipeline", "bootstrap"]



"""
Isolates interactive init prompt flow for event bridge.
"""

import click

from samcli.lib.schemas.cli_paginator import do_paginate_cli
from samcli.lib.schemas.schemas_api_caller import SchemasApiCaller
from samcli.lib.schemas.schemas_aws_config import get_aws_configuration_choice, get_schemas_client
from samcli.lib.schemas.schemas_cli_message_generator import (
    construct_cli_display_message_for_registries,
    construct_cli_display_message_for_schemas,
)
from samcli.lib.schemas.schemas_constants import (
    EVENT_BRIDGE_SOURCE,
    EVENT_BRIDGE_SOURCE_DETAIL_TYPE,
    PAGE_LIMIT,
    SCHEMA_NAME,
    SCHEMA_ROOT,
    SCHEMAS_REGISTRY,
)


def get_schema_template_details(schemas_api_caller):
    """
    Calls schemas APIs to fetch available selection and returns schema details based on user selection.
    :param schemas_api_caller:
    :return:
    """
    registry_name = _get_registry_cli_choice(schemas_api_caller)
    schema_full_name = _get_schema_cli_choice(schemas_api_caller, registry_name)
    schema_latest_version = schemas_api_caller.get_latest_schema_version(registry_name, schema_full_name)
    get_schema_metadata_response = schemas_api_caller.get_schema_metadata(registry_name, schema_full_name)
    return {
        "registry_name": registry_name,
        "schema_full_name": schema_full_name,
        "schema_version": schema_latest_version,
        "event_source": get_schema_metadata_response["event_source"],
        "event_source_detail_type": get_schema_metadata_response["event_source_detail_type"],
        "schema_root_name": get_schema_metadata_response["schema_root_name"],
        "schemas_package_hierarchy": get_schema_metadata_response["schemas_package_hierarchy"],
    }


def _get_registry_cli_choice(schemas_api_caller):
    """Returns registry choice if one registry is present otherwise prompt for selection"""
    registries = _fetch_available_registries(schemas_api_caller, dict(), None)
    registry_pages = registries["registry_pages"]
    # If only one registry don't prompt for choice
    if len(registry_pages) == 1 and len(registry_pages.get(0)) == 1:
        return registry_pages.get(0)[0]

    # more than one registries
    click.echo("Which Schema Registry would you like to use?")
    next_token = registries.get("next_token")
    is_last_page = next_token is None
    return _prompt_for_registry_choice(
        schemas_api_caller, registry_pages, 0, next_token, is_last_page, last_page_number=None
    )


def _prompt_for_registry_choice(
    schemas_api_caller, registry_pages, page_to_render, next_token, is_last_page, last_page_number
):
    # construct CLI message
    cli_display_message = construct_cli_display_message_for_registries(page_to_render + 1, last_page_number)

    # get customer decision
    cli_response = do_paginate_cli(registry_pages, page_to_render, PAGE_LIMIT, is_last_page, cli_display_message)

    # user selected item
    if cli_response.get("choice") is not None:
        return cli_response.get("choice")

    # user decided to paginate
    page_to_render = cli_response.get("page_to_render")
    if registry_pages.get(page_to_render) is None:
        registries = _fetch_available_registries(schemas_api_caller, registry_pages, next_token)
        registry_pages = registries["registry_pages"]
        next_token = registries.get("next_token")
        is_last_page = next_token is None
    if is_last_page and last_page_number is None:
        last_page_number = page_to_render + 1
    return _prompt_for_registry_choice(
        schemas_api_caller, registry_pages, page_to_render, next_token, is_last_page, last_page_number
    )


def _get_schema_cli_choice(schemas_api_caller, registry_name):
    """Returns registry registry choice if one registry is present otherwise prompt for  selection"""
    schemas = _fetch_available_schemas(schemas_api_caller, registry_name, dict(), None)
    schema_pages = schemas["schema_pages"]
    # If only one schema don't prompt for choice
    if len(schema_pages) == 1 and len(schema_pages.get(0)) == 1:
        return schema_pages.get(0)[0]

    # more than one schema
    click.echo("\nWhich Schema would you like to use?")
    next_token = schemas.get("next_token")
    is_last_page = next_token is None
    return _prompt_for_schemas_choice(
        schemas_api_caller, registry_name, schema_pages, 0, next_token, is_last_page, last_page_number=None
    )


def _prompt_for_schemas_choice(
    schemas_api_caller, registry_name, schema_pages, page_to_render, next_token, is_last_page, last_page_number
):
    # construct CLI message
    cli_display_message = construct_cli_display_message_for_schemas(page_to_render + 1, last_page_number)

    # get customer decision
    cli_response = do_paginate_cli(schema_pages, page_to_render, PAGE_LIMIT, is_last_page, cli_display_message)

    # user selected item
    if cli_response.get("choice") is not None:
        return cli_response["choice"]

    # user decided to paginate
    page_to_render = cli_response.get("page_to_render")
    if schema_pages.get(page_to_render) is None:
        schemas = _fetch_available_schemas(schemas_api_caller, registry_name, schema_pages, next_token)
        schema_pages = schemas["schema_pages"]
        next_token = schemas.get("next_token")
        is_last_page = next_token is None
    if is_last_page and last_page_number is None:
        last_page_number = page_to_render + 1
    return _prompt_for_schemas_choice(
        schemas_api_caller, registry_name, schema_pages, page_to_render, next_token, is_last_page, last_page_number
    )


def _fetch_available_schemas(schemas_api_caller, registry_name, schema_pages, next_token):
    """calls schemas api fetch schemas for given registry. Two CLI pages are fetched at a time."""
    list_schemas_response = schemas_api_caller.list_schemas(registry_name, next_token, PAGE_LIMIT)
    schemas = list_schemas_response["schemas"]

    # divided response into pages
    pages = _construct_cli_page(schemas, PAGE_LIMIT)
    for page in range(0, len(pages)):
        schema_pages.update({len(schema_pages): pages.get(page)})
    next_token = list_schemas_response.get("next_token")
    return {"schema_pages": schema_pages, "next_token": next_token}


def _fetch_available_registries(schemas_api_caller, registry_pages, next_token):
    """calls schemas api to fetch registries. Two CLI pages are fetched at a time."""
    list_registries_response = schemas_api_caller.list_registries(next_token, PAGE_LIMIT)
    registries = list_registries_response["registries"]

    # sort registries alphabetically by name
    registries.sort()

    # divided response into pages
    pages = _construct_cli_page(registries, PAGE_LIMIT)
    for page in range(0, len(pages)):
        registry_pages.update({len(registry_pages): pages.get(page)})
    next_token = list_registries_response.get("next_token")
    return {"registry_pages": registry_pages, "next_token": next_token}


def _construct_cli_page(items, item_per_page):
    """Responsible for splitting items into CLI pages.
    Currently CLI pages are list of dictionary [0:{0:s1, 1:s2: 3:s3}, 1: {4:s4, 5:s5: 6:s6}]
    We maintain the page detail and item index details."""
    pages = [
        items[i * item_per_page : (i + 1) * item_per_page]
        for i in range((len(items) + item_per_page - 1) // item_per_page)
    ]
    index = 0
    schema_dict = dict()
    for page in pages:
        schema_dict.update({index: page})
        index = index + 1

    return schema_dict


def get_schemas_template_parameter(schema_template_details):
    """Schemas cookiecutter template parameter mapping"""
    return {
        SCHEMAS_REGISTRY: schema_template_details["registry_name"],
        SCHEMA_NAME: schema_template_details["schema_root_name"],
        EVENT_BRIDGE_SOURCE: schema_template_details["event_source"],
        EVENT_BRIDGE_SOURCE_DETAIL_TYPE: schema_template_details["event_source_detail_type"],
        SCHEMA_ROOT: schema_template_details["schemas_package_hierarchy"],
    }


def get_schemas_api_caller():
    aws_configuration = get_aws_configuration_choice()
    schemas_client = get_schemas_client(aws_configuration["profile"], aws_configuration["region"])
    return SchemasApiCaller(schemas_client)



"""
Class used to parse and update template with new field
"""
import logging
from abc import abstractmethod
from copy import deepcopy
from typing import Any, List

from yaml.parser import ParserError

from samcli.yamlhelper import parse_yaml_file

LOG = logging.getLogger(__name__)


class TemplateModifier:
    def __init__(self, location):
        self.template_location = location
        self.template = self._get_template()
        self.copy_of_original_template = deepcopy(self.template)

    def modify_template(self):
        """
        This method modifies the template by first added the new field to the template
        and then run a sanity check on the template to know if the template matches the
        CFN yaml
        """
        self._update_template_fields()
        self._write(self.template)
        if not self._sanity_check():
            self._write(self.copy_of_original_template)

    @abstractmethod
    def _update_template_fields(self):
        pass

    def _section_position(self, section: str, position: int = 0) -> int:
        """
        validate if a section in the template exist

        Parameters
        ----------
        section : str
            A section in the SAM template
        position : int
            position to start searching for the section

        Returns
        -------
        int
            index of section in the template list
        """
        template = self.template[position:]
        for index, line in enumerate(template):
            if line.startswith(section):
                section_index = index + position if position else index
                return section_index
        return -1

    def _add_fields_to_section(self, position: int, fields: List[str]) -> Any:
        """
        Adds fields to section in the template

        Parameters
        ----------
        position : int
            position to start searching for the section
        fields : str
            fields to be added to the SAM template

        Returns
        -------
        list
            array with updated template data
        """
        section_space_count = self.template[position].find(self.template[position].strip())
        start_position = position + 1
        template = self.template[start_position:]
        for index, line in enumerate(template):
            if not (line.find(line.strip()) > section_space_count or line.startswith("#")):
                return self.template[: start_position + index] + fields + self.template[start_position + index :]
        return self.template

    def _field_position(self, position: int, field: str) -> Any:
        """
        Checks if the field needed to be added to the SAM template already exist in the template

        Parameters
        ----------
        position : int
            section position to start the search
        field : str
            Field name

        Returns
        -------
        int
            index of the field if it exist else -1
        """
        template = self.template[position:]
        for index, line in enumerate(template):
            if field in line:
                return position + index
            if not (line.startswith(" ") or line.startswith("#")):
                break
        return -1

    def _sanity_check(self) -> bool:
        """
        Conducts sanity check on template using yaml parser to ensure the updated template meets
        CFN template criteria

        Returns
        -------
        bool
            True if templates passes sanity check else False
        """
        try:
            parse_template = parse_yaml_file(self.template_location)
            return bool(parse_template)
        except ParserError:
            self._print_sanity_check_error()
            return False

    @abstractmethod
    def _print_sanity_check_error(self):
        pass

    @abstractmethod
    def _write(self, template: list):
        pass

    @abstractmethod
    def _get_template(self):
        pass



"""
Class to manage all the prompts during a guided sam deploy
"""

import logging
from typing import Any, Dict, List, Optional

import click
from click import confirm, prompt
from click.types import FuncParamType

from samcli.commands._utils.options import _space_separated_list_func_type
from samcli.commands._utils.template import (
    get_template_parameters,
)
from samcli.commands.deploy.auth_utils import auth_per_resource
from samcli.commands.deploy.code_signer_utils import (
    extract_profile_name_and_owner_from_existing,
    prompt_profile_name,
    prompt_profile_owner,
    signer_config_per_function,
)
from samcli.commands.deploy.exceptions import GuidedDeployFailedError
from samcli.commands.deploy.guided_config import GuidedConfig
from samcli.commands.deploy.utils import sanitize_parameter_overrides
from samcli.lib.bootstrap.bootstrap import manage_stack
from samcli.lib.bootstrap.companion_stack.companion_stack_manager import CompanionStackManager
from samcli.lib.config.samconfig import DEFAULT_CONFIG_FILE_NAME, DEFAULT_ENV
from samcli.lib.intrinsic_resolver.intrinsics_symbol_table import IntrinsicsSymbolTable
from samcli.lib.package.ecr_utils import is_ecr_url
from samcli.lib.package.image_utils import NoImageFoundException, NonLocalImageException, tag_translation
from samcli.lib.providers.provider import Function, ResourceIdentifier, Stack, get_resource_full_path_by_id
from samcli.lib.providers.sam_function_provider import SamFunctionProvider
from samcli.lib.providers.sam_stack_provider import SamLocalStackProvider
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.defaults import get_default_aws_region
from samcli.lib.utils.packagetype import IMAGE

LOG = logging.getLogger(__name__)


class GuidedContext:
    # pylint: disable=too-many-statements
    def __init__(
        self,
        template_file,
        stack_name,
        s3_bucket,
        image_repository,
        image_repositories,
        s3_prefix,
        region=None,
        profile=None,
        confirm_changeset=None,
        capabilities=None,
        signing_profiles=None,
        parameter_overrides=None,
        save_to_config=True,
        config_section=None,
        config_env=None,
        config_file=None,
        disable_rollback=None,
    ):
        self.template_file = template_file
        self.stack_name = stack_name
        self.s3_bucket = s3_bucket
        self.image_repository = image_repository
        self.image_repositories = image_repositories
        self.s3_prefix = s3_prefix
        self.region = region
        self.profile = profile
        self.confirm_changeset = confirm_changeset
        self.capabilities = (capabilities,)
        self.parameter_overrides_from_cmdline = parameter_overrides
        self.save_to_config = save_to_config
        self.config_section = config_section
        self.config_env = config_env
        self.config_file = config_file
        self.guided_stack_name = None
        self.guided_s3_bucket = None
        self.guided_image_repository = None
        self.guided_image_repositories = None
        self.guided_s3_prefix = None
        self.guided_region = None
        self.guided_profile = None
        self.signing_profiles = signing_profiles
        self._capabilities = None
        self._parameter_overrides = None
        self.start_bold = "\033[1m"
        self.end_bold = "\033[0m"
        self.color = Colored()
        self.function_provider = None
        self.disable_rollback = disable_rollback

    @property
    def guided_capabilities(self):
        return self._capabilities

    @property
    def guided_parameter_overrides(self):
        return self._parameter_overrides

    # pylint: disable=too-many-statements
    def guided_prompts(self, parameter_override_keys):
        """
        Start an interactive cli prompt to collection information for deployment

        Parameters
        ----------
        parameter_override_keys
            The keys of parameters to override, for each key, customers will be asked to provide a value
        """
        default_stack_name = self.stack_name or "sam-app"
        default_region = self.region or get_default_aws_region()
        default_capabilities = self.capabilities[0] or ("CAPABILITY_IAM",)
        default_config_env = self.config_env or DEFAULT_ENV
        default_config_file = self.config_file or DEFAULT_CONFIG_FILE_NAME
        input_capabilities = None
        config_env = None
        config_file = None

        click.echo(
            self.color.yellow(
                "\n\tSetting default arguments for 'sam deploy'\n\t========================================="
            )
        )

        stack_name = prompt(
            f"\t{self.start_bold}Stack Name{self.end_bold}", default=default_stack_name, type=click.STRING
        )
        region = prompt(f"\t{self.start_bold}AWS Region{self.end_bold}", default=default_region, type=click.STRING)
        global_parameter_overrides = {IntrinsicsSymbolTable.AWS_REGION: region}
        input_parameter_overrides = self.prompt_parameters(
            parameter_override_keys, self.parameter_overrides_from_cmdline, self.start_bold, self.end_bold
        )
        stacks, _ = SamLocalStackProvider.get_stacks(
            self.template_file,
            parameter_overrides=sanitize_parameter_overrides(input_parameter_overrides),
            global_parameter_overrides=global_parameter_overrides,
        )

        click.secho("\t#Shows you resources changes to be deployed and require a 'Y' to initiate deploy")
        confirm_changeset = confirm(
            f"\t{self.start_bold}Confirm changes before deploy{self.end_bold}", default=self.confirm_changeset
        )
        click.secho("\t#SAM needs permission to be able to create roles to connect to the resources in your template")
        capabilities_confirm = confirm(
            f"\t{self.start_bold}Allow SAM CLI IAM role creation{self.end_bold}", default=True
        )

        if not capabilities_confirm:
            input_capabilities = prompt(
                f"\t{self.start_bold}Capabilities{self.end_bold}",
                default=list(default_capabilities),
                type=FuncParamType(func=_space_separated_list_func_type),
            )

        click.secho("\t#Preserves the state of previously provisioned resources when an operation fails")
        disable_rollback = confirm(f"\t{self.start_bold}Disable rollback{self.end_bold}", default=self.disable_rollback)

        self.prompt_authorization(stacks)
        self.prompt_code_signing_settings(stacks)

        save_to_config = confirm(
            f"\t{self.start_bold}Save arguments to configuration file{self.end_bold}", default=True
        )
        if save_to_config:
            config_file = prompt(
                f"\t{self.start_bold}SAM configuration file{self.end_bold}",
                default=default_config_file,
                type=click.STRING,
            )
            config_env = prompt(
                f"\t{self.start_bold}SAM configuration environment{self.end_bold}",
                default=default_config_env,
                type=click.STRING,
            )

        click.echo("\n\tLooking for resources needed for deployment:")
        s3_bucket = manage_stack(profile=self.profile, region=region)
        click.secho(f"\n\tManaged S3 bucket: {s3_bucket}", bold=True)
        click.echo("\tA different default S3 bucket can be set in samconfig.toml")

        image_repositories = self.prompt_image_repository(
            stack_name, stacks, self.image_repositories, region, s3_bucket, self.s3_prefix
        )

        self.guided_stack_name = stack_name
        self.guided_s3_bucket = s3_bucket
        self.guided_image_repositories = image_repositories
        self.guided_s3_prefix = stack_name
        self.guided_region = region
        self.guided_profile = self.profile
        self._capabilities = input_capabilities if input_capabilities else default_capabilities
        self._parameter_overrides = (
            input_parameter_overrides if input_parameter_overrides else self.parameter_overrides_from_cmdline
        )
        self.save_to_config = save_to_config
        self.config_env = config_env if config_env else default_config_env
        self.config_file = config_file if config_file else default_config_file
        self.confirm_changeset = confirm_changeset
        self.disable_rollback = disable_rollback

    def prompt_authorization(self, stacks: List[Stack]):
        auth_required_per_resource = auth_per_resource(stacks)

        for resource, authorization_required in auth_required_per_resource:
            if not authorization_required:
                auth_confirm = confirm(
                    f"\t{self.start_bold}{resource} may not have authorization defined, Is this okay?{self.end_bold}",
                    default=False,
                )
                if not auth_confirm:
                    raise GuidedDeployFailedError(msg="Security Constraints Not Satisfied!")

    def prompt_code_signing_settings(self, stacks: List[Stack]):
        """
        Prompt code signing settings to ask whether customers want to code sign their code and
        display signing details.

        Parameters
        ----------
        stacks : List[Stack]
            List of stacks to search functions and layers
        """
        (functions_with_code_sign, layers_with_code_sign) = signer_config_per_function(stacks)

        # if no function or layer definition found with code signing, skip it
        if not functions_with_code_sign and not layers_with_code_sign:
            LOG.debug("No function or layer definition found with code sign config, skipping")
            return

        click.echo("\n\t#Found code signing configurations in your function definitions")
        sign_functions = confirm(
            f"\t{self.start_bold}Do you want to sign your code?{self.end_bold}",
            default=True,
        )

        if not sign_functions:
            LOG.debug("User skipped code signing, continuing rest of the process")
            self.signing_profiles = None
            return

        if not self.signing_profiles:
            self.signing_profiles = {}

        click.echo("\t#Please provide signing profile details for the following functions & layers")

        for function_name in functions_with_code_sign:
            (profile_name, profile_owner) = extract_profile_name_and_owner_from_existing(
                function_name, self.signing_profiles
            )

            click.echo(f"\t#Signing profile details for function '{function_name}'")
            profile_name = prompt_profile_name(profile_name, self.start_bold, self.end_bold)
            profile_owner = prompt_profile_owner(profile_owner, self.start_bold, self.end_bold)
            self.signing_profiles[function_name] = {"profile_name": profile_name, "profile_owner": profile_owner}
            self.signing_profiles[function_name]["profile_owner"] = "" if not profile_owner else profile_owner

        for layer_name, functions_use_this_layer in layers_with_code_sign.items():
            (profile_name, profile_owner) = extract_profile_name_and_owner_from_existing(
                layer_name, self.signing_profiles
            )
            click.echo(
                f"\t#Signing profile details for layer '{layer_name}', "
                f"which is used by functions {functions_use_this_layer}"
            )
            profile_name = prompt_profile_name(profile_name, self.start_bold, self.end_bold)
            profile_owner = prompt_profile_owner(profile_owner, self.start_bold, self.end_bold)
            self.signing_profiles[layer_name] = {"profile_name": profile_name, "profile_owner": profile_owner}
            self.signing_profiles[layer_name]["profile_owner"] = "" if not profile_owner else profile_owner

        LOG.debug("Signing profile names and owners %s", self.signing_profiles)

    def prompt_parameters(
        self, parameter_override_from_template, parameter_override_from_cmdline, start_bold, end_bold
    ):
        _prompted_param_overrides = {}
        if parameter_override_from_template:
            for parameter_key, parameter_properties in parameter_override_from_template.items():
                no_echo = parameter_properties.get("NoEcho", False)
                if no_echo:
                    parameter = prompt(
                        f"\t{start_bold}Parameter {parameter_key}{end_bold}", type=click.STRING, hide_input=True
                    )
                    _prompted_param_overrides[parameter_key] = {"Value": parameter, "Hidden": True}
                else:
                    parameter = prompt(
                        f"\t{start_bold}Parameter {parameter_key}{end_bold}",
                        default=_prompted_param_overrides.get(
                            parameter_key,
                            self._get_parameter_value(
                                parameter_key, parameter_properties, parameter_override_from_cmdline
                            ),
                        ),
                        type=click.STRING,
                    )
                    _prompted_param_overrides[parameter_key] = {"Value": parameter, "Hidden": False}
        return _prompted_param_overrides

    def prompt_image_repository(
        self,
        stack_name,
        stacks: List[Stack],
        image_repositories: Optional[Dict[str, str]],
        region: str,
        s3_bucket: str,
        s3_prefix: str,
    ) -> Dict[str, str]:
        """
        Prompt for the image repository to push the images.
        For each image function found in build artifacts, it will prompt for an image repository.

        Parameters
        ----------
        stack_name : List[Stack]
            Name of the stack to be deployed.

        stacks : List[Stack]
            List of stacks to look for image functions.

        image_repositories: Dict[str, str]
            Dictionary with function logical ID as key and image repo URI as value.

        region: str
            Region for the image repos.

        s3_bucket: str
            s3 bucket URI to be used for uploading companion stack template

        s3_prefix: str
            s3 prefix to be used for uploading companion stack template

        Returns
        -------
        Dict[str, str]
            A dictionary contains image function logical ID as key, image repository as value.
        """
        image_repositories = image_repositories if image_repositories is not None else {}
        updated_repositories = {}
        for image_repo_func_id, image_repo_uri in image_repositories.items():
            repo_full_path = get_resource_full_path_by_id(stacks, ResourceIdentifier(image_repo_func_id))
            if repo_full_path:
                updated_repositories[repo_full_path] = image_repo_uri
        self.function_provider = SamFunctionProvider(stacks, ignore_code_extraction_warnings=True)
        manager = CompanionStackManager(stack_name, region, s3_bucket, s3_prefix)

        function_logical_ids = [
            function.full_path for function in self.function_provider.get_all() if function.packagetype == IMAGE
        ]

        functions_without_repo = [
            function_logical_id
            for function_logical_id in function_logical_ids
            if function_logical_id not in updated_repositories
        ]

        manager.set_functions(function_logical_ids, updated_repositories)

        create_all_repos = self.prompt_create_all_repos(
            function_logical_ids, functions_without_repo, updated_repositories
        )
        if create_all_repos:
            updated_repositories.update(manager.get_repository_mapping())
        else:
            updated_repositories = self.prompt_specify_repos(functions_without_repo, updated_repositories)
            manager.set_functions(function_logical_ids, updated_repositories)

        updated_repositories = self.prompt_delete_unreferenced_repos(
            [manager.get_repo_uri(repo) for repo in manager.get_unreferenced_repos()], updated_repositories
        )
        GuidedContext.verify_images_exist_locally(self.function_provider.functions)

        manager.sync_repos()
        return updated_repositories

    def prompt_specify_repos(
        self,
        functions_without_repos: List[str],
        image_repositories: Dict[str, str],
    ) -> Dict[str, str]:
        """
        Show prompts for each function that isn't associated with a image repo

        Parameters
        ----------
        functions_without_repos: List[str]
            List of functions without associating repos

        image_repositories: Dict[str, str]
            Current image repo dictionary with function logical ID as key and image repo URI as value.

        Returns
        -------
        Dict[str, str]
            Updated image repo dictionary with values(image repo URIs) filled by user input
        """
        updated_repositories = image_repositories.copy()
        for function_logical_id in functions_without_repos:
            image_uri = prompt(
                f"\t {self.start_bold}ECR repository for {function_logical_id}{self.end_bold}",
                type=click.STRING,
            )
            if not is_ecr_url(image_uri):
                raise GuidedDeployFailedError(f"Invalid Image Repository ECR URI: {image_uri}")

            updated_repositories[function_logical_id] = image_uri

        return updated_repositories

    def prompt_create_all_repos(
        self, functions: List[str], functions_without_repo: List[str], existing_mapping: Dict[str, str]
    ) -> bool:
        """
        Prompt whether to create all repos

        Parameters
        ----------
        functions: List[str]
            List of function logical IDs that are image based

        functions_without_repo: List[str]
            List of function logical IDs that do not have an ECR image repo specified

        existing_mapping: Dict[str, str]
            Current image repo dictionary with function logical ID as key and image repo URI as value.
            This dict will be shown in the terminal.

        Returns
        -------
        Boolean
            Returns False if there is no missing function or denied by prompt
        """
        if not functions:
            return False

        # Case for when all functions do not have mapped repo
        if functions == functions_without_repo:
            click.echo("\t Image repositories: Not found.")
            click.echo(
                "\t #Managed repositories will be deleted when "
                "their functions are removed from the template and deployed"
            )
            return confirm(
                f"\t {self.start_bold}Create managed ECR repositories for all functions?{self.end_bold}", default=True
            )

        functions_with_repo_count = len(functions) - len(functions_without_repo)
        click.echo(
            "\t Image repositories: "
            f"Found ({functions_with_repo_count} of {len(functions)})"
            " #Different image repositories can be set in samconfig.toml"
        )
        for function_logical_id, repo_uri in existing_mapping.items():
            click.echo(f"\t {function_logical_id}: {repo_uri}")

        # Case for all functions do have mapped repo
        if not functions_without_repo:
            return False

        click.echo(
            "\t #Managed repositories will be deleted when their functions are "
            "removed from the template and deployed"
        )
        return (
            confirm(
                f"\t {self.start_bold}Create managed ECR repositories for the "
                f"{len(functions_without_repo)} functions without?{self.end_bold}",
                default=True,
            )
            if functions_without_repo
            else True
        )

    def prompt_delete_unreferenced_repos(
        self, unreferenced_repo_uris: List[str], image_repositories: Dict[str, str]
    ) -> Dict[str, str]:
        """
        Prompt user for deleting unreferenced companion stack image repos.
        Throws GuidedDeployFailedError if delete repos has been denied by the user.
        This function does not actually remove the functions from the stack.

        Parameters
        ----------

        unreferenced_repo_uris: List[str]
            List of unreferenced image repos that need to be deleted.
        image_repositories: Dict[str, str]
            Dictionary of image repo URIs with key as function logical ID and value as image repo URI

        Returns
        -------
        Dict[str, str]
            Copy of image_repositories that have unreferenced image repos removed
        """
        output_image_repositories = image_repositories.copy()

        if not unreferenced_repo_uris:
            return output_image_repositories

        click.echo("\t Checking for unreferenced ECR repositories to clean-up: " f"{len(unreferenced_repo_uris)} found")
        for repo_uri in unreferenced_repo_uris:
            click.echo(f"\t  {repo_uri}")
        delete_repos = confirm(
            f"\t {self.start_bold}Delete the unreferenced repositories listed above when deploying?{self.end_bold}",
            default=False,
        )
        if not delete_repos:
            click.echo("\t Deployment aborted!")
            click.echo(
                "\t #The deployment was aborted to prevent "
                "unreferenced managed ECR repositories from being deleted.\n"
                "\t #You may remove repositories from the SAMCLI "
                "managed stack to retain them and resolve this unreferenced check."
            )
            raise GuidedDeployFailedError("Unreferenced Auto Created ECR Repos Must Be Deleted.")

        for function_logical_id, repo_uri in image_repositories.items():
            if repo_uri in unreferenced_repo_uris:
                del output_image_repositories[function_logical_id]
                break
        return output_image_repositories

    @staticmethod
    def verify_images_exist_locally(functions: Dict[str, Function]) -> None:
        """
        Verify all images associated with deploying functions exist locally.

        Parameters
        ----------
        functions: Dict[str, Function]
            Dictionary of functions in the stack to be deployed with key as their logical ID.
        """
        for _, function_prop in functions.items():
            if function_prop.packagetype != IMAGE:
                continue
            image = function_prop.imageuri
            try:
                tag_translation(image)
            except NonLocalImageException:
                LOG.debug("Image URI is not pointing to local. Skipping verification.")
            except NoImageFoundException as ex:
                raise GuidedDeployFailedError("No images found to deploy, try running sam build") from ex

    def run(self):

        try:
            _parameter_override_keys = get_template_parameters(template_file=self.template_file)
        except ValueError as ex:
            LOG.debug("Failed to parse SAM template", exc_info=ex)
            raise GuidedDeployFailedError(str(ex)) from ex

        guided_config = GuidedConfig(template_file=self.template_file, section=self.config_section)
        guided_config.read_config_showcase(
            self.config_file or DEFAULT_CONFIG_FILE_NAME,
        )

        self.guided_prompts(_parameter_override_keys)

        if self.save_to_config:
            guided_config.save_config(
                self._parameter_overrides,
                self.config_env or DEFAULT_ENV,
                self.config_file or DEFAULT_CONFIG_FILE_NAME,
                stack_name=self.guided_stack_name,
                s3_bucket=self.guided_s3_bucket,
                s3_prefix=self.guided_s3_prefix,
                image_repositories=self.guided_image_repositories,
                region=self.guided_region,
                profile=self.guided_profile,
                confirm_changeset=self.confirm_changeset,
                capabilities=self._capabilities,
                signing_profiles=self.signing_profiles,
                disable_rollback=self.disable_rollback,
            )

    @staticmethod
    def _get_parameter_value(
        parameter_key: str, parameter_properties: Dict, parameter_override_from_cmdline: Dict
    ) -> Any:
        """
        This function provide the value of a parameter. If the command line/config file have "override_parameter"
        whose key exist in the template file parameters, it will use the corresponding value.
        Otherwise, it will use its default value in template file.

        :param parameter_key: key of parameter
        :param parameter_properties: properties of that parameters from template file
        :param parameter_override_from_cmdline: parameter_override from command line/config file
        """
        if parameter_override_from_cmdline and parameter_override_from_cmdline.get(parameter_key, None):
            return parameter_override_from_cmdline[parameter_key]
        # Make sure the default is casted to a string.
        return str(parameter_properties.get("Default", ""))



"""
Custom Click options for hook name
"""

import logging
import os

import click

from samcli.cli.context import Context
from samcli.cli.global_config import GlobalConfig
from samcli.commands._utils.constants import DEFAULT_BUILT_TEMPLATE_PATH
from samcli.commands._utils.experimental import (
    ExperimentalFlag,
    prompt_experimental,
    set_experimental,
    update_experimental_context,
)
from samcli.lib.hook.exceptions import InvalidHookWrapperException
from samcli.lib.hook.hook_wrapper import IacHookWrapper, get_available_hook_packages_ids

LOG = logging.getLogger(__name__)


class HookNameOption(click.Option):
    """
    A custom option class that allows do custom validation for the SAM CLI commands options in case if hook name
    is passed. It also calls the correct IaC prepare hook, and update the SAM CLI commands options based on the
    prepare hook output.
    """

    def __init__(self, *args, **kwargs):
        self.hook_name_option_name = "hook_name"
        self._force_prepare = kwargs.pop("force_prepare", True)
        self._invalid_coexist_options = kwargs.pop("invalid_coexist_options", [])
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        opt_name = self.hook_name_option_name.replace("_", "-")
        if self.hook_name_option_name not in opts and self.hook_name_option_name not in ctx.default_map:
            return super().handle_parse_result(ctx, opts, args)
        command_name = ctx.command.name
        if command_name in ["invoke", "start-lambda", "start-api"]:
            command_name = f"local {command_name}"
        # validate the hook_name value exists
        hook_name = opts.get(self.hook_name_option_name) or ctx.default_map[self.hook_name_option_name]
        iac_hook_wrapper = None
        try:
            iac_hook_wrapper = IacHookWrapper(hook_name)
        except InvalidHookWrapperException as e:
            raise click.BadParameter(
                f"{hook_name} is not a valid hook name."
                f"{os.linesep}valid package ids: {get_available_hook_packages_ids()}"
            ) from e

        self._validate_coexist_options(opt_name, opts)

        _validate_build_command_parameters(command_name, opts)

        if not _check_experimental_flag(hook_name, command_name, opts, ctx.default_map):
            return super().handle_parse_result(ctx, opts, args)

        try:
            self._call_prepare_hook(iac_hook_wrapper, opts)
        except Exception as ex:
            # capture exceptions from prepare hook to emit in track_command
            c = Context.get_current_context()
            c.exception = ex

        return super().handle_parse_result(ctx, opts, args)

    def _call_prepare_hook(self, iac_hook_wrapper, opts):
        # call prepare hook
        built_template_path = DEFAULT_BUILT_TEMPLATE_PATH
        if not self._force_prepare and os.path.exists(built_template_path):
            LOG.info("Skipped prepare hook. Current application is already prepared.")
        else:
            LOG.info("Running Prepare Hook to prepare the current application")

            iac_project_path = os.getcwd()
            output_dir_path = os.path.join(iac_project_path, ".aws-sam-iacs", "iacs_metadata")
            if not os.path.exists(output_dir_path):
                os.makedirs(output_dir_path, exist_ok=True)

            debug = opts.get("debug", False)
            aws_profile = opts.get("profile")
            aws_region = opts.get("region")
            skip_prepare_infra = opts.get("skip_prepare_infra", False)

            metadata_file = iac_hook_wrapper.prepare(
                output_dir_path, iac_project_path, debug, aws_profile, aws_region, skip_prepare_infra
            )

            LOG.info("Prepare hook completed and metadata file generated at: %s", metadata_file)
            opts["template_file"] = metadata_file

    def _validate_coexist_options(self, opt_name, opts):
        # validate coexist options
        for invalid_opt in self._invalid_coexist_options:
            invalid_opt_name = invalid_opt.replace("-", "_")
            if invalid_opt_name in opts:
                raise click.BadParameter(
                    f"Parameters {opt_name}, and {','.join(self._invalid_coexist_options)} cannot be used together"
                )


def _validate_build_command_parameters(command_name, opts):
    # validate build-image is provided in case of build using container
    # add this validation here to avoid running hook prepare and there is issue
    if command_name == "build" and opts.get("use_container") and not opts.get("build_image"):
        raise click.UsageError("Missing required parameter --build-image.")


def _check_experimental_flag(hook_name, command_name, opts, default_map):
    # check beta-feature
    experimental_entry = ExperimentalFlag.IaCsSupport.get(hook_name)
    beta_features = _get_customer_input_beta_features_option(default_map, experimental_entry, opts)

    # check if beta feature flag is required for a specific hook package
    # The IaCs support experimental flag map will contain only the beta IaCs. In case we support the external
    # hooks, we need to first know that the hook package is an external, and to handle the beta feature of it
    # using different approach
    if beta_features is None and experimental_entry is not None:
        iac_support_message = _get_iac_support_experimental_prompt_message(hook_name, command_name)
        if not prompt_experimental(experimental_entry, iac_support_message):
            LOG.debug("Experimental flag is disabled and prepare hook is not run")
            return False
    elif not beta_features:
        LOG.debug("--beta-features flag is disabled and prepare hook is not run")
        return False
    elif beta_features:
        LOG.debug("--beta-features flag is enabled, enabling experimental flag.")
        set_experimental(experimental_entry, True)
        update_experimental_context()
    return True


def _get_customer_input_beta_features_option(default_map, experimental_entry, opts):

    # Get the beta-features flag value from the command parameters if provided.
    beta_features = opts.get("beta_features")
    if beta_features is not None:
        return beta_features

    # Get the beta-features flag value from the SamConfig toml file if provided.
    beta_features = default_map.get("beta_features")
    if beta_features is not None:
        return beta_features

    # Get the beta-features flag value from the environment variables.
    if experimental_entry:
        gc = GlobalConfig()
        beta_features = gc.get_value(experimental_entry, default=None, value_type=bool, is_flag=True)
        if beta_features is not None:
            return beta_features
        return gc.get_value(ExperimentalFlag.All, default=None, value_type=bool, is_flag=True)

    return None


def _get_iac_support_experimental_prompt_message(hook_name: str, command: str) -> str:
    """
    return the customer prompt message for a specific hook package.

    Parameters
    ----------
    hook_name: str
        the hook name to determine what is the supported iac

    command: str
        the current sam command
    Returns
    -------
    str
        the customer prompt message for a specific IaC.
    """

    supported_iacs_messages = {
        "terraform": (
            "Supporting Terraform applications is a beta feature.\n"
            "Please confirm if you would like to proceed using AWS SAM CLI with terraform application.\n"
            f"You can also enable this beta feature with 'sam {command} --beta-features'."
        )
    }
    return supported_iacs_messages.get(hook_name, "")



"""
CLI command for "bootstrap", which sets up a SAM development environment
"""
import click

from samcli.cli.main import aws_creds_options, common_options, pass_context
from samcli.lib.bootstrap import bootstrap
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

SHORT_HELP = "Set up development environment for AWS SAM applications."

HELP_TEXT = """
Sets up a development environment for AWS SAM applications.

Currently this creates, if one does not exist, a managed S3 bucket for your account in your working AWS region.
"""


@click.command("bootstrap", short_help=SHORT_HELP, help=HELP_TEXT, context_settings=dict(max_content_width=120))
@common_options
@aws_creds_options
@pass_context
@track_command
@check_newer_version
def cli(ctx):
    do_cli(ctx.region, ctx.profile)  # pragma: no cover


def do_cli(region, profile):  # pragma: no cover
    bucket_name = bootstrap.manage_stack(profile=profile, region=region)
    click.echo("Source Bucket: " + bucket_name)



"""
Build Related Exceptions.
"""


class UnsupportedBuilderLibraryVersionError(Exception):
    def __init__(self, container_name: str, error_msg: str) -> None:
        msg = (
            "You are running an outdated version of Docker container '{container_name}' that is not compatible with"
            "this version of SAM CLI. Please upgrade to continue to continue with build. Reason: '{error_msg}'"
        )
        Exception.__init__(self, msg.format(container_name=container_name, error_msg=error_msg))


class ContainerBuildNotSupported(Exception):
    pass


class BuildError(Exception):
    def __init__(self, wrapped_from: str, msg: str) -> None:
        self.wrapped_from = wrapped_from
        Exception.__init__(self, msg)


class BuildInsideContainerError(Exception):
    pass


class DockerConnectionError(BuildError):
    def __init__(self, msg: str) -> None:
        BuildError.__init__(self, "DockerConnectionError", msg)


class DockerfileOutSideOfContext(BuildError):
    def __init__(self, msg: str) -> None:
        BuildError.__init__(self, "DockerfileOutSideOfContext", msg)


class DockerBuildFailed(BuildError):
    def __init__(self, msg: str) -> None:
        BuildError.__init__(self, "DockerBuildFailed", msg)


class MissingBuildMethodException(BuildError):
    def __init__(self, msg: str) -> None:
        BuildError.__init__(self, "MissingBuildMethodException", msg)


class InvalidBuildGraphException(Exception):
    def __init__(self, msg: str) -> None:
        Exception.__init__(self, msg)



"""
Isolates interactive init prompt flow. Expected to call generator logic at end of flow.
"""
import logging
import pathlib
import tempfile
from typing import Optional, Tuple

import click
from botocore.exceptions import ClientError, WaiterError

from samcli.commands._utils.options import generate_next_command_recommendation
from samcli.commands.exceptions import InvalidInitOptionException, SchemasApiException
from samcli.commands.init.init_flow_helpers import (
    _get_image_from_runtime,
    _get_runtime_from_image,
    _get_templates_with_dependency_manager,
    get_architectures,
    get_sorted_runtimes,
)
from samcli.commands.init.init_generator import do_generate
from samcli.commands.init.init_templates import InitTemplates, InvalidInitTemplateError
from samcli.commands.init.interactive_event_bridge_flow import (
    get_schema_template_details,
    get_schemas_api_caller,
    get_schemas_template_parameter,
)
from samcli.lib.config.samconfig import DEFAULT_CONFIG_FILE_NAME
from samcli.lib.schemas.schemas_code_manager import do_download_source_code_binding, do_extract_and_merge_schemas_code
from samcli.lib.utils.osutils import remove
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.local.common.runtime_template import (
    LAMBDA_IMAGES_RUNTIMES_MAP,
    get_provided_runtime_from_custom_runtime,
    is_custom_runtime,
)

LOG = logging.getLogger(__name__)


# pylint: disable=too-many-arguments
def do_interactive(
    location,
    pt_explicit,
    package_type,
    runtime,
    architecture,
    base_image,
    dependency_manager,
    output_dir,
    name,
    app_template,
    no_input,
    tracing,
    application_insights,
):
    """
    Implementation of the ``cli`` method when --interactive is provided.
    It will ask customers a few questions to init a template.
    """
    if app_template:
        location_opt_choice = "1"
    else:
        click.echo("Which template source would you like to use?")
        click.echo("\t1 - AWS Quick Start Templates\n\t2 - Custom Template Location")
        location_opt_choice = click.prompt("Choice", type=click.Choice(["1", "2"]), show_choices=False)

    generate_application(
        location,
        pt_explicit,
        package_type,
        runtime,
        architecture,
        base_image,
        dependency_manager,
        output_dir,
        name,
        app_template,
        no_input,
        location_opt_choice,
        tracing,
        application_insights,
    )


def generate_application(
    location,
    pt_explicit,
    package_type,
    runtime,
    architecture,
    base_image,
    dependency_manager,
    output_dir,
    name,
    app_template,
    no_input,
    location_opt_choice,
    tracing,
    application_insights,
):  # pylint: disable=too-many-arguments
    """
    The method holds the decision logic for generating an application
    Parameters
    ----------
    location : str
        Location to SAM template
    pt_explicit : bool
        boolean representing if the customer explicitly stated packageType
    package_type : str
        Zip or Image
    runtime : str
        AWS Lambda runtime or Custom runtime
    architecture : str
        The architecture type 'x86_64' and 'arm64' in AWS
    base_image : str
        AWS Lambda base image
    dependency_manager : str
        Runtime's Dependency manager
    output_dir : str
        Project output directory
    name : str
        name of the project
    app_template : str
        AWS Serverless Application template
    no_input : bool
        Whether to prompt for input or to accept default values
        (the default is False, which prompts the user for values it doesn't know for baking)
    location_opt_choice : int
        User input for selecting how to get customer a vended serverless application
    tracing : bool
        boolen value to determine if X-Ray tracing show be activated or not
    application_insights : bool
        boolean value to determine if AppInsights monitoring should be enabled or not
    """
    if location_opt_choice == "1":
        _generate_from_use_case(
            location,
            pt_explicit,
            package_type,
            runtime,
            base_image,
            dependency_manager,
            output_dir,
            name,
            app_template,
            architecture,
            tracing,
            application_insights,
        )

    else:
        _generate_from_location(
            location,
            package_type,
            runtime,
            dependency_manager,
            output_dir,
            name,
            no_input,
            tracing,
            application_insights,
        )


# pylint: disable=too-many-statements
def _generate_from_location(
    location, package_type, runtime, dependency_manager, output_dir, name, no_input, tracing, application_insights
):
    location = click.prompt("\nTemplate location (git, mercurial, http(s), zip, path)", type=str)
    summary_msg = """
-----------------------
Generating application:
-----------------------
Location: {location}
Output Directory: {output_dir}
    """.format(
        location=location, output_dir=output_dir
    )
    click.echo(summary_msg)
    do_generate(
        location,
        package_type,
        runtime,
        dependency_manager,
        output_dir,
        name,
        no_input,
        None,
        tracing,
        application_insights,
    )


# pylint: disable=too-many-statements
def _generate_from_use_case(
    location: Optional[str],
    pt_explicit: bool,
    package_type: Optional[str],
    runtime: Optional[str],
    base_image: Optional[str],
    dependency_manager: Optional[str],
    output_dir: Optional[str],
    name: Optional[str],
    app_template: Optional[str],
    architecture: Optional[str],
    tracing: Optional[bool],
    application_insights: Optional[bool],
) -> None:
    templates = InitTemplates()
    runtime_or_base_image = runtime if runtime else base_image
    package_type_filter_value = package_type if pt_explicit else None
    preprocessed_options = templates.get_preprocessed_manifest(
        runtime_or_base_image, app_template, package_type_filter_value, dependency_manager
    )
    question = "Choose an AWS Quick Start application template"
    use_case = _get_choice_from_options(
        None,
        preprocessed_options,
        question,
        "Template",
    )

    default_app_template_properties = _generate_default_hello_world_application(
        use_case, package_type, runtime, base_image, dependency_manager, pt_explicit
    )

    chosen_app_template_properties = _get_app_template_properties(
        preprocessed_options, use_case, base_image, default_app_template_properties
    )
    runtime, base_image, package_type, dependency_manager, template_chosen = chosen_app_template_properties

    if tracing is None:
        tracing = prompt_user_to_enable_tracing()

    if application_insights is None:
        application_insights = prompt_user_to_enable_application_insights()

    app_template = template_chosen["appTemplate"]
    base_image = (
        LAMBDA_IMAGES_RUNTIMES_MAP.get(str(runtime)) if not base_image and package_type == IMAGE else base_image
    )

    if not name:
        name = click.prompt("\nProject name", type=str, default="sam-app")

    location = templates.location_from_app_template(package_type, runtime, base_image, dependency_manager, app_template)

    final_architecture = get_architectures(architecture)
    lambda_supported_runtime = (
        get_provided_runtime_from_custom_runtime(runtime) if is_custom_runtime(runtime) else runtime
    )
    extra_context = {
        "project_name": name,
        "runtime": lambda_supported_runtime,
        "architectures": {"value": final_architecture},
    }

    # executing event_bridge logic if call is for Schema dynamic template
    is_dynamic_schemas_template = templates.is_dynamic_schemas_template(
        package_type, app_template, runtime, base_image, dependency_manager
    )
    if is_dynamic_schemas_template:
        schemas_api_caller = get_schemas_api_caller()
        schema_template_details = _get_schema_template_details(schemas_api_caller)
        schemas_template_parameter = get_schemas_template_parameter(schema_template_details)
        extra_context = {**schemas_template_parameter, **extra_context}

    no_input = True
    summary_msg = generate_summary_message(
        package_type, runtime, base_image, dependency_manager, output_dir, name, app_template, final_architecture
    )

    click.echo(summary_msg)
    command_suggestions = generate_next_command_recommendation(
        [
            ("Create pipeline", f"cd {name} && sam pipeline init --bootstrap"),
            ("Validate SAM template", f"cd {name} && sam validate"),
            ("Test Function in the Cloud", f"cd {name} && sam sync --stack-name {{stack-name}} --watch"),
        ]
    )
    click.secho(command_suggestions, fg="yellow")

    do_generate(
        location,
        package_type,
        lambda_supported_runtime,
        dependency_manager,
        output_dir,
        name,
        no_input,
        extra_context,
        tracing,
        application_insights,
    )
    # executing event_bridge logic if call is for Schema dynamic template
    if is_dynamic_schemas_template:
        _package_schemas_code(
            lambda_supported_runtime, schemas_api_caller, schema_template_details, output_dir, name, location
        )


def _generate_default_hello_world_application(
    use_case: str,
    package_type: Optional[str],
    runtime: Optional[str],
    base_image: Optional[str],
    dependency_manager: Optional[str],
    pt_explicit: bool,
) -> Tuple:
    """
    Generate the default Hello World template if Hello World Example is selected

    Parameters
    ----------
    use_case : str
        Type of template example selected
    package_type : Optional[str]
        The package type, 'Zip' or 'Image', see samcli/lib/utils/packagetype.py
    runtime : Optional[str]
        AWS Lambda function runtime
    base_image : Optional[str]
        AWS Lambda function base-image
    dependency_manager : Optional[str]
        dependency manager
    pt_explicit : bool
        True --package-type was passed or Vice versa

    Returns
    -------
    Tuple
        configuration for a default Hello World Example
    """
    is_package_type_image = bool(package_type == IMAGE)
    if use_case == "Hello World Example" and not (runtime or base_image or is_package_type_image or dependency_manager):
        if click.confirm("\nUse the most popular runtime and package type? (Python and zip)"):
            runtime, package_type, dependency_manager, pt_explicit = "python3.9", ZIP, "pip", True
    return (runtime, package_type, dependency_manager, pt_explicit)


def _get_app_template_properties(
    preprocessed_options: dict, use_case: str, base_image: Optional[str], template_properties: Tuple
) -> Tuple:
    """
    This is the heart of the interactive flow, this method fetchs the templates options needed to generate a template

    Parameters
    ----------
    preprocessed_options : dict
        Preprocessed manifest from https://github.com/aws/aws-sam-cli-app-templates
    use_case : Optional[str]
        Type of template example selected
    base_image : str
        AWS Lambda function base-image
    template_properties : Tuple
        Tuple of template properties like runtime, packages type and dependency manager

    Returns
    -------
    Tuple
        Tuple of template configuration and the chosen template

    Raises
    ------
    InvalidInitOptionException
        exception raised when invalid option is provided
    """
    runtime, package_type, dependency_manager, pt_explicit = template_properties
    runtime_options = preprocessed_options[use_case]
    runtime = None if is_custom_runtime(runtime) else runtime
    if not runtime and not base_image:
        question = "Which runtime would you like to use?"
        runtime = _get_choice_from_options(runtime, runtime_options, question, "Runtime")

    if base_image:
        runtime = _get_runtime_from_image(base_image)

    package_types_options = runtime_options.get(runtime)
    if not package_types_options:
        raise InvalidInitOptionException(f"Lambda Runtime {runtime} is not supported for {use_case} examples.")
    if not pt_explicit:
        message = "What package type would you like to use?"
        package_type = _get_choice_from_options(None, package_types_options, message, "Package type")
        if package_type == IMAGE:
            base_image = _get_image_from_runtime(runtime)

    dependency_manager_options = package_types_options.get(package_type)
    if not dependency_manager_options:
        raise InvalidInitOptionException(
            f"{package_type} package type is not supported for {use_case} examples and runtime {runtime} selected."
        )

    dependency_manager = _get_dependency_manager(dependency_manager_options, dependency_manager, runtime)
    template_chosen = _get_app_template_choice(dependency_manager_options, dependency_manager)
    return (runtime, base_image, package_type, dependency_manager, template_chosen)


def prompt_user_to_enable_tracing():
    """
    Prompt user to if X-Ray Tracing should activated for functions in the SAM template and vice versa
    """
    if click.confirm("\nWould you like to enable X-Ray tracing on the function(s) in your application? "):
        doc_link = "https://aws.amazon.com/xray/pricing/"
        click.echo(f"X-Ray will incur an additional cost. View {doc_link} for more details")
        return True
    return False


def prompt_user_to_enable_application_insights():
    """
    Prompt user to choose if AppInsights monitoring should be enabled for their application and vice versa
    """
    doc_link = "https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/cloudwatch-application-insights.html"
    if click.confirm(
        f"\nWould you like to enable monitoring using CloudWatch Application Insights?"
        f"\nFor more info, please view {doc_link}"
    ):
        pricing_link = (
            "https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring"
            "/appinsights-what-is.html#appinsights-pricing"
        )
        click.echo(f"AppInsights monitoring may incur additional cost. View {pricing_link} for more details")
        return True
    return False


def _get_choice_from_options(chosen, options, question, msg):

    if chosen:
        return chosen

    click_choices = []

    options_list = options if isinstance(options, list) else list(options.keys())
    options_list = get_sorted_runtimes(options_list) if msg == "Runtime" else options_list

    if not options_list:
        raise InvalidInitOptionException(f"There are no {msg} options available to be selected.")

    if len(options_list) == 1:
        click.echo(
            f"\nBased on your selections, the only {msg} available is {options_list[0]}."
            + f"\nWe will proceed to selecting the {msg} as {options_list[0]}."
        )
        return options_list[0]

    click.echo(f"\n{question}")

    for index, option in enumerate(options_list):
        click.echo(f"\t{index+1} - {option}")
        click_choices.append(str(index + 1))
    choice = click.prompt(msg, type=click.Choice(click_choices), show_choices=False)
    return options_list[int(choice) - 1]


def _get_app_template_choice(templates_options, dependency_manager):
    templates = _get_templates_with_dependency_manager(templates_options, dependency_manager)
    chosen_template = templates[0]
    if len(templates) > 1:
        click.echo("\nSelect your starter template")
        click_template_choices = []
        for index, template in enumerate(templates):
            click.echo(f"\t{index+1} - {template['displayName']}")
            click_template_choices.append(str(index + 1))
        template_choice = click.prompt("Template", type=click.Choice(click_template_choices), show_choices=False)
        chosen_template = templates[int(template_choice) - 1]
    return chosen_template


def _get_dependency_manager(options, dependency_manager, runtime):
    valid_dep_managers = sorted(list(set(template["dependencyManager"] for template in options)))
    if not dependency_manager:
        if len(valid_dep_managers) == 1:
            dependency_manager = valid_dep_managers[0]
            click.echo(
                f"\nBased on your selections, the only dependency manager available is {dependency_manager}."
                + f"\nWe will proceed copying the template using {dependency_manager}."
            )
        else:
            question = "Which dependency manager would you like to use?"
            dependency_manager = _get_choice_from_options(
                dependency_manager, valid_dep_managers, question, "Dependency manager"
            )
    elif dependency_manager and dependency_manager not in valid_dep_managers:
        msg = (
            f"Lambda Runtime {runtime} and dependency manager {dependency_manager} "
            + "do not have an available initialization template."
        )
        raise InvalidInitTemplateError(msg)
    return dependency_manager


def _get_schema_template_details(schemas_api_caller):
    try:
        return get_schema_template_details(schemas_api_caller)
    except ClientError as e:
        raise SchemasApiException(
            "Exception occurs while getting Schemas template parameter. %s" % e.response["Error"]["Message"]
        ) from e


def _package_schemas_code(runtime, schemas_api_caller, schema_template_details, output_dir, name, location):
    try:
        click.echo("Trying to get package schema code")
        with tempfile.NamedTemporaryFile(delete=False) as download_location:
            do_download_source_code_binding(runtime, schema_template_details, schemas_api_caller, download_location)
            do_extract_and_merge_schemas_code(download_location, output_dir, name, location)
    except (ClientError, WaiterError) as e:
        raise SchemasApiException(
            "Exception occurs while packaging Schemas code. %s" % e.response["Error"]["Message"]
        ) from e
    finally:
        remove(download_location.name)


def generate_summary_message(
    package_type, runtime, base_image, dependency_manager, output_dir, name, app_template, architecture
):
    """
    Parameters
    ----------
    package_type : str
        The package type, 'Zip' or 'Image', see samcli/lib/utils/packagetype.py
    runtime : str
        AWS Lambda function runtime
    base_image : str
        base image
    dependency_manager : str
        dependency manager
    output_dir : str
        the directory where project will be generated in
    name : str
        Project Name
    app_template : str
        application template generated
    architecture : list
        Architecture type either x86_64 or arm64 on AWS lambda

    Returns
    -------
    str
        Summary Message of the application template generated
    """

    summary_msg = ""
    if package_type == ZIP:
        summary_msg = f"""
    -----------------------
    Generating application:
    -----------------------
    Name: {name}
    Runtime: {runtime}
    Architectures: {architecture[0]}
    Dependency Manager: {dependency_manager}
    Application Template: {app_template}
    Output Directory: {output_dir}
    Configuration file: {pathlib.Path(output_dir).joinpath(name, DEFAULT_CONFIG_FILE_NAME)}
    
    Next steps can be found in the README file at {pathlib.Path(output_dir).joinpath(name, "README.md")}
        """
    elif package_type == IMAGE:
        summary_msg = f"""
    -----------------------
    Generating application:
    -----------------------
    Name: {name}
    Base Image: {base_image}
    Architectures: {architecture[0]}
    Dependency Manager: {dependency_manager}
    Output Directory: {output_dir}
    Configuration file: {pathlib.Path(output_dir).joinpath(name, DEFAULT_CONFIG_FILE_NAME)}

    Next steps can be found in the README file at {pathlib.Path(output_dir).joinpath(name, "README.md")}
    """

    return summary_msg



"""
Initialize an arbitrary project
"""

import functools
import logging
import shutil

from cookiecutter import config, exceptions, repository

from samcli.lib.utils import osutils

from .exceptions import ArbitraryProjectDownloadFailed

LOG = logging.getLogger(__name__)


BAD_LOCATION_ERROR_MSG = (
    "Please verify your location. The following types of location are supported:"
    "\n\n* Github: gh:user/repo (or) https://github.com/user/repo (or) git@github.com:user/repo.git"
    "\n          For Git repositories, you must use location of the root of the repository."
    "\n\n* Mercurial: hg+ssh://hg@bitbucket.org/repo"
    "\n\n* Http(s): https://example.com/code.zip"
    "\n\n* Local Path: /path/to/code.zip"
)


def generate_non_cookiecutter_project(location, output_dir):
    """
    Uses Cookiecutter APIs to download a project at given ``location`` to the ``output_dir``.
    This does *not* run cookiecutter on the downloaded project.

    Parameters
    ----------
    location : str
        Path to where the project is. This supports all formats of location cookiecutter supports
        (ex: zip, git, ssh, hg, local zipfile)

        NOTE: This value *cannot* be a local directory. We didn't see a value in simply copying the directory
        contents to ``output_dir`` without any processing.

    output_dir : str
        Directory where the project should be downloaded to

    Returns
    -------
    str
        Name of the directory where the project was downloaded to.

    Raises
    ------
    cookiecutter.exception.CookiecutterException if download failed for some reason
    """

    LOG.debug("Downloading project from %s to %s", location, output_dir)

    # Don't prompt ever
    no_input = True

    # Expand abbreviations in URL such as gh:awslabs/aws-sam-cli
    location = repository.expand_abbreviations(location, config.BUILTIN_ABBREVIATIONS)

    # If this is a zip file, download and unzip into output directory
    if repository.is_zip_file(location):
        LOG.debug("%s location is a zip file", location)
        download_fn = functools.partial(
            repository.unzip, zip_uri=location, is_url=repository.is_repo_url(location), no_input=no_input
        )

    # Else, treat it as a git/hg/ssh URL and try to clone
    elif repository.is_repo_url(location):
        LOG.debug("%s location is a source control repository", location)
        download_fn = functools.partial(repository.clone, repo_url=location, no_input=no_input)

    else:
        raise ArbitraryProjectDownloadFailed(msg=BAD_LOCATION_ERROR_MSG)

    try:
        return _download_and_copy(download_fn, output_dir)
    except exceptions.RepositoryNotFound as ex:
        # Download failed because the zip or the repository was not found
        raise ArbitraryProjectDownloadFailed(msg=BAD_LOCATION_ERROR_MSG) from ex


def _download_and_copy(download_fn, output_dir):
    """
    Runs the download function to download files into a temporary directory and then copy the files over to
    the ``output_dir``

    Parameters
    ----------
    download_fn : function
        Method to be called to download. It needs to accept a parameter called `clone_to_dir`. This will be
        set to the temporary directory

    output_dir : str
        Path to the directory where files will be copied to

    Returns
    -------
    output_dir
    """

    with osutils.mkdir_temp(ignore_errors=True) as tempdir:
        downloaded_dir = download_fn(clone_to_dir=tempdir)
        osutils.copytree(downloaded_dir, output_dir, ignore=shutil.ignore_patterns("*.git"))

    return output_dir



""" Manage Git repo """

import logging
import os
import platform
import shutil
import subprocess
from pathlib import Path

# import check_output alone so that it can be patched without affecting
# other parts of subprocess.
from subprocess import check_output
from typing import Optional

from samcli.lib.utils import osutils
from samcli.lib.utils.osutils import rmtree_callback

LOG = logging.getLogger(__name__)


class CloneRepoException(Exception):
    """
    Exception class when clone repo fails.
    """


class CloneRepoUnstableStateException(CloneRepoException):
    """
    Exception class when clone repo enters an unstable state.
    """


class ManifestNotFoundException(Exception):
    """
    Exception class when request Manifest file return 404.
    """


class GitRepo:
    """
    Class for managing a Git repo, currently it has a clone functionality only

    Attributes
    ----------
    url: str
        The URL of this Git repository, example "https://github.com/aws/aws-sam-cli"
    local_path: Path
        The path of the last local clone of this Git repository. Can be used in conjunction with clone_attempted
        to avoid unnecessary multiple cloning of the repository.
    clone_attempted: bool
        whether an attempt to clone this Git repository took place or not. Can be used in conjunction with local_path
        to avoid unnecessary multiple cloning of the repository

    Methods
    -------
    clone(self, clone_dir: Path, clone_name, replace_existing=False) -> Path:
        creates a local clone of this Git repository. (more details in the method documentation).
    """

    def __init__(self, url: str) -> None:
        self.url: str = url
        self.local_path: Optional[Path] = None
        self.clone_attempted: bool = False

    @staticmethod
    def _ensure_clone_directory_exists(clone_dir: Path) -> None:
        try:
            clone_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
        except OSError as ex:
            LOG.warning("WARN: Unable to create clone directory.", exc_info=ex)
            raise

    @staticmethod
    def _git_executable() -> str:
        if platform.system().lower() == "windows":
            executables = ["git", "git.cmd", "git.exe", "git.bat"]
        else:
            executables = ["git"]

        for executable in executables:
            try:
                with subprocess.Popen([executable], stdout=subprocess.PIPE, stderr=subprocess.PIPE):
                    # No exception. Let's pick this
                    return executable
            except OSError as ex:
                LOG.debug("Unable to find executable %s", executable, exc_info=ex)

        raise OSError("Cannot find git, was looking at executables: {}".format(executables))

    def clone(self, clone_dir: Path, clone_name: str, replace_existing: bool = False, commit: str = "") -> Path:
        """
        creates a local clone of this Git repository.
        This method is different from the standard Git clone in the following:
        1. It accepts the path to clone into as a clone_dir (the parent directory to clone in) and a clone_name (The
           name of the local folder) instead of accepting the full path (the join of both) in one parameter
        2. It removes the "*.git" files/directories so the clone is not a GitRepo any more
        3. It has the option to replace the local folder(destination) if already exists

        Parameters
        ----------
        clone_dir: Path
            The directory to create the local clone inside
        clone_name: str
            The dirname of the local clone
        replace_existing: bool
            Whether to replace the current local clone directory if already exists or not
        commit: str
            if a commit is provided, it will checkout out the commit in the clone repo
        Returns
        -------
            The path of the created local clone

        Raises
        ------
        OSError:
            when file management errors like unable to mkdir, copytree, rmtree ...etc
        CloneRepoException:
            General errors like for example; if an error occurred while running `git clone`
            or if the local_clone already exists and replace_existing is not set
        CloneRepoUnstableStateException:
            when reaching unstable state, for example with replace_existing flag set, unstable state can happen
            if removed the current local clone but failed to copy the new one from the temp location to the destination
        """

        GitRepo._ensure_clone_directory_exists(clone_dir=clone_dir)
        # clone to temp then move to the destination(repo_local_path)
        with osutils.mkdir_temp(ignore_errors=True) as tempdir:
            try:
                temp_path = os.path.normpath(os.path.join(tempdir, clone_name))
                git_executable: str = GitRepo._git_executable()
                LOG.info("\nCloning from %s (process may take a moment)", self.url)
                command = [git_executable, "clone", self.url, clone_name]
                if platform.system().lower() == "windows":
                    LOG.debug(
                        "Configure core.longpaths=true in git clone. "
                        "You might also need to enable long paths in Windows registry."
                    )
                    command += ["--config", "core.longpaths=true"]
                check_output(
                    command,
                    cwd=tempdir,
                    stderr=subprocess.STDOUT,
                )

                # bind a certain sam cli release to a specific commit of the aws-sam-cli-app-templates's repo, avoiding
                # regression
                if commit:
                    self._checkout_commit(temp_path, commit)

                self.local_path = self._persist_local_repo(temp_path, clone_dir, clone_name, replace_existing)
                return self.local_path
            except OSError as ex:
                LOG.warning("WARN: Could not clone repo %s", self.url, exc_info=ex)
                raise
            except subprocess.CalledProcessError as clone_error:
                output = clone_error.output.decode("utf-8")
                if "not found" in output.lower():
                    LOG.warning("WARN: Could not clone repo %s", self.url, exc_info=clone_error)
                raise CloneRepoException(output) from clone_error
            finally:
                self.clone_attempted = True

    @staticmethod
    def _persist_local_repo(temp_path: str, dest_dir: Path, dest_name: str, replace_existing: bool) -> Path:
        dest_path = os.path.normpath(dest_dir.joinpath(dest_name))
        try:
            if Path(dest_path).exists():
                if not replace_existing:
                    raise CloneRepoException(f"Can not clone to {dest_path}, directory already exist")
                LOG.debug("Removing old repo at %s", dest_path)
                shutil.rmtree(dest_path, onerror=rmtree_callback)

            LOG.debug("Copying from %s to %s", temp_path, dest_path)
            # Todo consider not removing the .git files/directories
            shutil.copytree(temp_path, dest_path, ignore=shutil.ignore_patterns("*.git"))
            return Path(dest_path)
        except (OSError, shutil.Error) as ex:
            # UNSTABLE STATE
            # it's difficult to see how this scenario could happen except weird permissions, user will need to debug
            msg = (
                "Unstable state when updating repo. "
                f"Check that you have permissions to create/delete files in {dest_dir} directory "
                "or file an issue at https://github.com/aws/aws-sam-cli/issues"
            )

            if platform.system().lower() == "windows":
                msg = (
                    "Failed to modify a local file when cloning app templates. "
                    "MAX_PATH should be enabled in the Windows registry."
                    "\nFor more details on how to enable MAX_PATH for Windows, please visit: "
                    "https://docs.aws.amazon.com/serverless-application-model/latest/"
                    "developerguide/install-sam-cli.html"
                )

            raise CloneRepoUnstableStateException(msg) from ex

    @staticmethod
    def _checkout_commit(repo_dir: str, commit: str):
        try:
            # if the checkout commit failed, it will use the latest commit instead
            git_executable = GitRepo._git_executable()
            check_output(
                [git_executable, "checkout", commit],
                cwd=repo_dir,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as checkout_error:
            output = checkout_error.output.decode("utf-8")
            if "fatal" in output.lower() or "error" in output.lower():
                LOG.warning("WARN: Commit not exist: %s, using the latest one", commit, exc_info=checkout_error)
